/**CHeaderFile*************************************************************
  PackageName [bdd]
  FileName    [bdd.h]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]

  Revision    [$Revision: 2.10 $]
  Date        [$Date: 2005/02/11 14:25:32 $]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]
  Description [The file bdd.h contains declaration of all external
               data structures neded for representation and
               manipulation of boolean functions with ROBDDs.]
  SeeAlso     [bddInt.h, bddMain.c ...]
  Copyright   [Copyright (C) 2000-2005 Robert Meolic
               This program is free software; you can redistribute it
               and/or modify it under the terms of the GNU General Public
               License (version 2) as published by the Free Software
               Foundation.

               This program is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 59 Temple Place, Suite 330,
               Boston, MA  02111-1307 USA.]

  ************************************************************************/

#ifndef _BDD
#define _BDD

#define BDD_RELEASE "2.3"

#include <pthread.h>
#include <sched.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include <unistd.h>
#include <math.h>

/* ON WIN32 THERE HAS TO BE DEFINED ESTWIN32 */
/* ON UNIX THERE HAS TO BE DEFINED ESTUNIX */

#ifdef ESTUNIX
#  ifndef EXTERN
#    define EXTERN extern
#  endif
#endif

#ifdef ESTWIN32
#ifndef TCL
#    ifdef BUILD_BDD
#      undef EXTERN
#      define EXTERN __declspec (dllexport)
#    else
#      ifdef USE_BDD
#        undef EXTERN
#        define EXTERN __declspec (dllimport)
#      else
#        define EXTERN extern
#      endif
#    endif
#else
#    ifdef BUILD_BDD
#      undef TCL_STORAGE_CLASS
#      define TCL_STORAGE_CLASS DLLEXPORT
#    else
#      undef TCL_STORAGE_CLASS
#      define TCL_STORAGE_CLASS DLLIMPORT
#    endif
#endif
#endif

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

#define FALSE 0
#define TRUE 1

/*-----------------------------------------------------------------------*/
/* Macro declarations                                                    */
/*-----------------------------------------------------------------------*/

#define Bdd_Cmp(f,g) (((f).p == (g).p) && ((f).mark == (g).mark))

/*-----------------------------------------------------------------------*/
/* Type declarations                                                     */
/*-----------------------------------------------------------------------*/

#ifndef _COMMONTYPE
#define _COMMONTYPE
typedef void *pvoid;
#endif

typedef char Bdd_Boolean;
typedef char *Bdd_String;
typedef FILE *Bdd_File;
typedef void (*Bdd_UserFunction)();

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

/* EDGE */
typedef struct Bdd_Edge {
  struct Bdd_Node *p;
  Bdd_Boolean mark;
} Bdd_Edge;

/* NODE */
typedef struct Bdd_Node {
  struct Bdd_Node *prev, *next; /* !MUST BE: 'prev'=first, 'next'=second */
  struct Bdd_Node *new;
  Bdd_Edge f, t;
  struct Bdd_Variable *v;
  int count;
} Bdd_Node;

/* VARIABLE */
typedef struct Bdd_Variable {
  struct Bdd_Variable *l, *r;      /* (BDD) left and right successor     */
  struct Bdd_Variable *parent;     /* (BDD) for maintaining symbol tree  */
  int colour;                      /* (BDD) for maintaining symbol tree  */
  Bdd_Edge term;                   /* (BDD) pointer to terminal          */
  int order;                       /* (BDD) variable ordering            */
  Bdd_String name;                 /* (BDD) name of variable             */
  Bdd_Boolean value;               /* (BDD) value = 0,1                  */
  int weight;                      /* (***) weight = 1,2,4... (decoding) */
} Bdd_Variable;

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

EXTERN int bdd_status;

EXTERN int bdd_counter;

EXTERN unsigned int bdd_maxNodes;

EXTERN unsigned int bdd_IteCacheSize;

EXTERN unsigned int bdd_EAxCacheSize;

EXTERN unsigned int bdd_RelOpCacheSize;

EXTERN unsigned int bdd_RelOpComplexCacheSize;

EXTERN Bdd_Edge bdd_termTrue;

EXTERN Bdd_Edge bdd_termFalse;

EXTERN Bdd_Edge bdd_termNull;

EXTERN Bdd_Edge bdd_undefiniedEdge;

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Function prototypes                                                   */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM bddMain.c                                     */
/*-----------------------------------------------------------------------*/

EXTERN void Bdd_InitPkg();

EXTERN void Bdd_ExitPkg();

EXTERN void Bdd_AboutPkg();

EXTERN Bdd_Edge Bdd_FindTerminal(Bdd_String x);

EXTERN Bdd_Edge Bdd_FoaTerminal(Bdd_String x);

EXTERN Bdd_Edge Bdd_FoaNode(Bdd_Variable *v, Bdd_Edge pf, Bdd_Edge pt);

EXTERN Bdd_Edge Bdd_NOT(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_ITE(Bdd_Edge f, Bdd_Edge g, Bdd_Edge h);

EXTERN Bdd_Edge Bdd_Restrict(Bdd_Edge f, Bdd_Variable *v, Bdd_Boolean value);

EXTERN Bdd_Edge Bdd_Compose(Bdd_Edge f, Bdd_Variable *v, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_E(Bdd_Edge f, Bdd_Variable *v);

EXTERN Bdd_Edge Bdd_A(Bdd_Edge f, Bdd_Variable *v);

EXTERN void Bdd_IncCounter();

EXTERN void Bdd_Fresh(Bdd_Node *f);

EXTERN void Bdd_Fortify(Bdd_Node *f);

EXTERN void Bdd_Garbage();

EXTERN Bdd_Edge Bdd_TransferMark(Bdd_Edge f, Bdd_Boolean mark);

EXTERN void Bdd_Exchange(Bdd_Edge *f, Bdd_Edge *g);

EXTERN void Bdd_ExchangeNeg(Bdd_Edge *f, Bdd_Edge *g);

EXTERN Bdd_Variable *Bdd_FindByOrder(Bdd_Variable *v, int n);

EXTERN Bdd_Boolean Bdd_FindFormula(Bdd_String x, Bdd_Edge *f);

EXTERN void Bdd_AddFormula(Bdd_String x, Bdd_Edge f);

EXTERN void Bdd_SaveFormula(Bdd_String prefix, Bdd_String name, Bdd_Edge f);

EXTERN void Bdd_AddCache(Bdd_UserFunction gc);

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM bddExpr.c                                     */
/*-----------------------------------------------------------------------*/

EXTERN Bdd_Edge Bdd_Expr(Bdd_String s, Bdd_String r);

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM bddOutput.c                                   */
/*-----------------------------------------------------------------------*/

EXTERN void Bdd_WriteFunction(Bdd_Edge f);

EXTERN void Bdd_WriteBDD(Bdd_Edge f);

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM bddStat.c                                     */
/*-----------------------------------------------------------------------*/

EXTERN void Bdd_SystemStat();

EXTERN void Bdd_FunctionStat();

EXTERN int Bdd_NodeNumber(Bdd_Node *p);

EXTERN int Bdd_NodeMaxLevel(Bdd_Node *p);

EXTERN float Bdd_NodeAvgLevel(Bdd_Node *p);

EXTERN int Bdd_ListUsed();

EXTERN int Bdd_ListMaxLength();

EXTERN float Bdd_ListAvgLength();

EXTERN int Bdd_CacheIteOk();

EXTERN int Bdd_CacheRelOpOk();

EXTERN int Bdd_CacheRelOpComplexOk();

/*-----------------------------------------------------------------------*/
/* FUNCTIONS EXPORTED FROM bddRel.c                                      */
/*-----------------------------------------------------------------------*/

EXTERN Bdd_Edge Bdd_RelOp(Bdd_Edge f, Bdd_Edge g,
                          Bdd_String format, Bdd_Boolean firsttime);

EXTERN Bdd_Edge Bdd_RelOpComplex(Bdd_Edge f, Bdd_Edge g,
                                 Bdd_String format, Bdd_Boolean firsttime);

EXTERN Bdd_Edge Bdd_ExA(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_ExR(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_ExS(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_ExQ(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_ExP(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_AxA(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_AxR(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_AxS(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_AxP(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_ExR_1(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_AxR_1(Bdd_Edge f, Bdd_Edge g);

EXTERN Bdd_Edge Bdd_R2S(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_S2R(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_P2Q(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_S2P(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_R2P(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_R2Q(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_R2P_S2Q(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_R2S_P2Q(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_R2P_P2R(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_S2R_Q2P(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_NewState(Bdd_Edge f);

EXTERN Bdd_Edge Bdd_NewState1(Bdd_Edge f, Bdd_String name);

EXTERN Bdd_Edge Bdd_NewState2(
                  Bdd_Edge f, Bdd_String name1, Bdd_String name2);

EXTERN Bdd_Edge Bdd_NewEq(Bdd_Edge f, Bdd_Edge g);

/**AutomaticEnd***********************************************************/

#endif  /* _BDD */
