/**CFile*******************************************************************
  PackageName [bdd]
  FileName    [bddExpr.c]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]

  Revision    [$Revision: 2.2 $]
  Date        [$Date: 2004/07/22 14:32:04 $]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]
  Description [The file bddExpr.c implements function Bdd_Expr which
               enables the usage of prefixed boolean expressions.]
  SeeAlso     [bdd.h, bddInt.h]
  Copyright   [Copyright (C) 2000-2004 Robert Meolic
               This program is free software; you can redistribute it
               and/or modify it under the terms of the GNU General Public
               License (version 2) as published by the Free Software
               Foundation.

               This program is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 59 Temple Place, Suite 330,
               Boston, MA  02111-1307 USA.]
  ************************************************************************/

#include "bddInt.h"

#define oAND 1
#define oOR 2
#define oEXOR 3

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static Bdd_Edge Izraz(Bdd_String s, int *i, Bdd_String *ch);

static void NextCh(Bdd_String s, int *i, Bdd_String *ch);

static Bdd_Edge Izrazi(Bdd_String s, int *i, Bdd_String *ch, Bdd_Edge g,
                        int op);

static int Op(Bdd_String s, int *i, Bdd_String *ch);

static Bdd_Boolean DoberZnak(char c);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function Bdd_Expr]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_Expr(Bdd_String s, Bdd_String r)
{
  Bdd_Edge pom;
  Bdd_String ch;
  int i;

  ch = (Bdd_String) malloc(255);

  i = 0;
  NextCh(s,&i,&ch);
  pom = Izraz(s,&i,&ch);

  free(ch);
  if (strcmp(r,"-")) Bdd_AddFormula(r,pom);
  return pom;
}

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function Izraz]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
Izraz(Bdd_String s, int *i, Bdd_String *ch)
{
  Bdd_Edge f;
  int op;

  if (!strcmp(*ch,"(")) {
    NextCh(s,i,ch);

    if (!strcmp(*ch, "NOT")) {
      NextCh(s,i,ch);
      f = Bdd_NOT(Izraz(s,i,ch));
    } else {
      if ((op = Op(s,i,ch))) {
	f = Izrazi(s, i, ch, Izraz(s,i,ch), op);
      } else {
	f = Izraz(s,i,ch);
      }
    }

    if (!strcmp(*ch, ")")) NextCh(s,i,ch);

    return f;
  }

  if (DoberZnak(*ch[0])) {
    if (!Bdd_FindFormula(*ch, &f)) f = Bdd_FoaTerminal(*ch);
    NextCh(s,i,ch);
    return f;
  }

  return bdd_undefiniedEdge;
}

/**Function****************************************************************
  Synopsis    [Function Izrazi]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Edge
Izrazi(Bdd_String s, int *i, Bdd_String *ch, Bdd_Edge g, int op)
{
  Bdd_Edge f, h;

  h = Izraz(s,i,ch);
  if (!Bdd_Cmp(h, bdd_undefiniedEdge)) {
    switch (op) {
	    case oAND:  f = Bdd_ITE(g, h, bdd_termFalse);
	                break;
	    case oOR:  f = Bdd_ITE(g, bdd_termTrue, h);
	                break;
	    case oEXOR:  f = Bdd_ITE(g, Bdd_NOT(h), h);
    }
    return Izrazi(s, i, ch, f, op);
  }
  return g;
}

/**Function****************************************************************
  Synopsis    [Function NextCh]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static void
NextCh(Bdd_String s, int *i, Bdd_String *ch)
{
  char c;
  int j;

    while (s[*i] == ' ' || s[*i] == '\t') (*i)++;
    j = *i;
    c = s[*i];

    if (c == '(') {
      (*i)++;
      strcpy(*ch,"(");
      return;
    }

    if (c == ')') {
      (*i)++;
      strcpy(*ch,")");
      return;
    }

    if (DoberZnak(c)) {
      (*i)++;
      while (DoberZnak(s[*i])) (*i)++;
      c = s[*i];

      if ( (c == ' ') || (c == '\t') || (c == '(') || (c == ')') ) {
        strncpy(*ch, &(s[j]), *i-j);
        (*ch)[*i-j] = 0;
        return;
      }
    }

    strcpy(*ch,"");
    return;
}

/**Function****************************************************************
  Synopsis    [Function Op]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
Op(Bdd_String s, int *i, Bdd_String *ch)
{
  if (!strcmp(*ch, "AND")) {
    NextCh(s,i,ch);
    return oAND;
  }

  if (!strcmp(*ch, "OR")) {
    NextCh(s,i,ch);
    return oOR;
  }

  if (!strcmp(*ch, "EXOR")) {
    NextCh(s,i,ch);
    return oEXOR;
  }

  return 0;
}

/**Function****************************************************************
  Synopsis    [Function DoberZnak]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static Bdd_Boolean
DoberZnak(char c)
{
  return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || 
         (c >= '0' && c <= '9') || c == '_' || c == '[' || c == ']' ||
         c == '$' || c == '.' || c == '&';
}

/**Function****************************************************************
  Synopsis    [Function UpCase]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_String
UpCase(Bdd_String s)
  {
    Bdd_String p;

    p = s;
    while (p[0])
      {
	p[0] = toupper(p[0]);
	p = &p[1];
      }
    return s;
  }

