/**CFile*******************************************************************
  PackageName [bdd]
  FileName    [bddMain.c]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]

  Revision    [$Revision: 2.7 $]
  Date        [$Date: 2005/02/11 14:25:32 $]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]
  Description [The file bddMain.c contains main functions for
               representation and manipulation of boolean functions
               with ROBDDs.]
  SeeAlso     [bdd.h, bddInt.h]
  Copyright   [Copyright (C) 2000-2004 Robert Meolic
               This program is free software; you can redistribute it
               and/or modify it under the terms of the GNU General Public
               License (version 2) as published by the Free Software
               Foundation.

               This program is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 59 Temple Place, Suite 330,
               Boston, MA  02111-1307 USA.]
  ************************************************************************/

#include "bddInt.h"

#ifdef ESTUNIX
#include <sys/resource.h>   /* used for setrlimit(), UNIX only */
#endif

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

/* Size of node and cache tables should be prime, e.g.:            */
/*   1009, 5003, 10007, 15013, 20021, 25013, 40009, 50021, 100043, */
/*   125003, 150001, 200003, 300007, 500057, 750019, 1000003 ...   */

#define SMALL_NODE_TABLE  50021
#define MEDIUM_NODE_TABLE 250007
#define BIG_NODE_TABLE    1250003
#define HUGE_NODE_TABLE   1600033

#define SMALL_ITE_CACHE  25013
#define MEDIUM_ITE_CACHE 100043
#define BIG_ITE_CACHE    500057
#define HUGE_ITE_CACHE   1250003

#define SMALL_EAX_CACHE  1009
#define MEDIUM_EAX_CACHE 10007
#define BIG_EAX_CACHE    100043
#define HUGE_EAX_CACHE   500057

#define SMALL_RELOP_CACHE  1009
#define MEDIUM_RELOP_CACHE 10007
#define BIG_RELOP_CACHE    100043
#define HUGE_RELOP_CACHE   500057

#define SMALL_RELOPCOMPLEX_CACHE  1009
#define MEDIUM_RELOPCOMPLEX_CACHE 10007
#define BIG_RELOPCOMPLEX_CACHE    100043
#define HUGE_RELOPCOMPLEX_CACHE   500057

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/* EXPORTED VARIABLES */

int bdd_status = 0;                             /* initialization status */

int bdd_counter;                                     /* formulae counter */

unsigned int bdd_maxNodes = SMALL_NODE_TABLE;
unsigned int bdd_IteCacheSize = BIG_ITE_CACHE;
unsigned int bdd_EAxCacheSize = BIG_EAX_CACHE;
unsigned int bdd_RelOpCacheSize = HUGE_RELOP_CACHE;
unsigned int bdd_RelOpComplexCacheSize = HUGE_RELOPCOMPLEX_CACHE;

Bdd_Edge bdd_termTrue;                                  /* terminal node */
Bdd_Edge bdd_termFalse;                                 /* terminal node */
Bdd_Edge bdd_termNull = {NULL, FALSE};                      /* null edge */
Bdd_Edge bdd_undefiniedEdge;                        /* supplemental edge */

/* INTERNAL VARIABLES */

Bdd_Edge debuggingEdge;                             /* supplemental edge */
Bdd_Variable *bddSymbolTree;                              /* symbol Tree */
BddNodeTable bddNodeTable;                                 /* node table */
BddFormula *bddFormulaeTree;                            /* formulae Tree */
BddCache *bddCacheList;                                    /* cache list */

Bdd_Node *bddNewNodes;                    /* used for garbage collection */
Bdd_Node *bddFirstNewNode;                /* used for garbage collection */
Bdd_Node *bddFreshNodes;                  /* used for garbage collection */
Bdd_Node *bddFreeNodes;                   /* used for garbage collection */

BddIteCacheTable bddIteCache;                               /* ITE cache */
BddEAxCacheTable bddEAxCache;                                /* EX cache */
BddRelOpCacheTable bddRelOpCache;                            /* AX cache */
BddRelOpComplexCacheTable bddRelOpComplexCache;              /* AX cache */

/**AutomaticStart*********************************************************/
/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

#ifndef TCLOUT
   void Bdd_Init(void) {}
#endif

/**Function****************************************************************
  Synopsis    [Function Bdd_InitPkg initialize BDD package.]
  Description [Bdd_InitPkg creates node table (bddNodeTable), symbol tree
               (bddSymbolTree), formulae tree (bddFormulaeTree), cache list
               (bddCacheList), ITE cache (bddIteCache), EX cache
               (bddExCache), AX cache (bddAxCache), terminal nodes
               (bdd_termTrue, bdd_termFalse), and auxiliary edges
               (bdd_termNull...). Bdd_InitPkg also initialize automatic
               garbage collection. BDD package must be initialized before
               any operation on BDDs is performed.]
  SideEffects [Allocate a lot of memory. If function call fails,
               decrease bdd_maxNodes.]
  SeeAlso     [Bdd_ExitPkg]
  ************************************************************************/

void
Bdd_InitPkg()
{
    int i;

#ifdef ESTUNIX
    struct rlimit rl;
#endif

    if (bdd_status == 1) {
      printf("\nBDD package is already initialized. ");
      return;
    } else {
      bdd_status = 1;
    }

#ifdef ESTUNIX

    rl.rlim_max  = rl.rlim_cur = RLIM_INFINITY;
    if (setrlimit(RLIMIT_STACK, &rl)) {
      fprintf(stderr,"ERROR: setrlimit\n");
      perror("setrlimit");
      exit(1);
    }

    /* TESTING */
    /*
    system("ulimit -a");
    system("sleep 5");
    */

#endif

    /*
    printf("\n\nSize of structures [bytes]:\n");
    printf("Bdd_Node = %d\n",sizeof(Bdd_Node));
    printf("Bdd_Edge = %d\n",sizeof(Bdd_Edge));
    printf("BddIteCache = %d\n",sizeof(BddIteCache));
    printf("BddEAxCache = %d\n",sizeof(BddEAxCache));
    printf("BddRelOpCache = %d\n",sizeof(BddRelOpCache));
    printf("BddRelOpComplexCache = %d\n\n",sizeof(BddRelOpComplexCache));
    */

    /* INITIALIZATION OF UNIQUE TABLE */
    bddNodeTable.size = bdd_maxNodes;
    bddNodeTable.table =
      (Bdd_Node **) calloc(bddNodeTable.size+1, sizeof(Bdd_Node *));

    /* INITIALIZATION OF SYMBOL TREE */
    bddSymbolTree =
      (Bdd_Variable *) malloc(sizeof(Bdd_Variable));
    bddSymbolTree->l = NULL;
    bddSymbolTree->r = NULL;
    bddSymbolTree->parent = NULL;
    bddSymbolTree->name = strdup("1");
    bddSymbolTree->value = TRUE;
    bddSymbolTree->order = 0;

    /* ADD TERMINAL EDGE TO UNIQUE TABLE */
    bddNodeTable.table[bddNodeTable.size] =
      (Bdd_Node *) calloc(1, sizeof(Bdd_Node));
    bddNodeTable.table[bddNodeTable.size]->v = bddSymbolTree;

    /* INITIALIZATION OF UNIQUE TABLE - CONTINUED */
    bddNodeTable.num = 1;
    bddNodeTable.nf = 0;
    bddNodeTable.foa = 0;
    bddNodeTable.add = 1;

    /* MAKE TERMINAL EDGES */
    bdd_termTrue.p = bddNodeTable.table[bddNodeTable.size];
    bdd_termTrue.p->f.p = NULL;
    bdd_termTrue.p->t.p = NULL;
    bdd_termTrue.mark = FALSE;
    bdd_termFalse = bdd_termTrue;
    bdd_termFalse.mark = TRUE;

    /* INITIALIZATION OF FORMULAE TREE */
    bddFormulaeTree = NULL;
    Bdd_AddFormula("termTrue",bdd_termTrue);
    Bdd_AddFormula("termFalse",bdd_termFalse);
    Bdd_AddFormula("1",bdd_termTrue);
    Bdd_AddFormula("0",bdd_termFalse);

    /* INITIALIZATION OF GARBAGE COLLECTION */
    bddFreeNodes =
      (Bdd_Node *) malloc((bdd_maxNodes) * sizeof(Bdd_Node));
    for (i=0; i<bdd_maxNodes-1; i++)
      bddFreeNodes[i].new = &bddFreeNodes[i+1];
    bddFreeNodes[bdd_maxNodes-1].new = NULL;
    bddNewNodes = bddFreshNodes = bddFirstNewNode = bdd_termTrue.p;
    bddNodeTable.generated = bdd_maxNodes;
    bdd_counter = 1;

    /* MAKE SUPPLEMENTAL EDGES */
    /* NOTE: Bdd_FoaTerminal requires initialized garbage collection! */
    bdd_undefiniedEdge = Bdd_FoaTerminal("`undefinied");
    debuggingEdge = bdd_undefiniedEdge;
    bddNodeTable.foa--; /* delete from statistic */
    bddNodeTable.add--; /* delete from statistic */

    /* INITIALIZATION OF NODE TABLE - CONTINUED */
    bddNodeTable.num = 1;
    bddNodeTable.max = 1;
    bddNodeTable.compare = 0;
    bddNodeTable.garbage = 0;

    /* INITIALIZATION OF CACHE LIST*/
    bddCacheList = NULL;

    /* INITIALIZATION OF ITE CACHE - USED FOR ITE OPERATION */
    bddIteCache.size = bdd_IteCacheSize;
    bddIteCache.search = bddIteCache.find = bddIteCache.overwrite = 0;
    if (!(bddIteCache.table = (BddIteCache *)
	  calloc(bddIteCache.size, sizeof(BddIteCache)))) {
      fprintf(stdout,"Bdd_InitPkg (ITE): Out of memoy!\n");
      exit(1);
    }
    Bdd_AddCache(BddIteGarbage);

    /* INITIALIZATION OF EAX CACHE - USED FOR QUANTIFICATIONS */
    bddEAxCache.size = bdd_EAxCacheSize;
    bddEAxCache.search = bddEAxCache.find = bddEAxCache.overwrite = 0;
    if (!(bddEAxCache.table = (BddEAxCache *)
	  calloc(bddEAxCache.size,sizeof(BddEAxCache)))) {
      fprintf(stdout,"Bdd_InitPkg (EAx): Out of memoy!\n");
      exit(1);
    }
    Bdd_AddCache(BddEAxGarbage);

    /* INITIALIZATION OF RelOp CACHE - USED FOR RELATIONAL OPERATIONS */
    bddRelOpCache.size = bdd_RelOpCacheSize;
    bddRelOpCache.search = 0;
    bddRelOpCache.find = 0;
    bddRelOpCache.write = 0;
    bddRelOpCache.overwrite = 0;
    if (!(bddRelOpCache.table = (BddRelOpCache *)
          calloc(bddRelOpCache.size,sizeof(BddRelOpCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOp): Out of memoy!\n");
      exit(1);
    }
    Bdd_AddCache(BddRelOpGarbage);

    /* INITIALIZATION OF RelOpComplex CACHE - ADVANCED RELATIONAL OPERAT. */
    bddRelOpComplexCache.size = bdd_RelOpComplexCacheSize;
    bddRelOpCache.search = 0;;
    bddRelOpComplexCache.search = 0;
    bddRelOpComplexCache.find = 0;
    bddRelOpComplexCache.write = 0;
    bddRelOpComplexCache.overwrite = 0;
    if (!(bddRelOpComplexCache.table = (BddRelOpComplexCache *)
          calloc(bddRelOpComplexCache.size, sizeof(BddRelOpComplexCache)))) {
      fprintf(stdout,"Bdd_InitPkg (RelOpComplex): Out of memoy!\n");
      exit(1);
    }
    Bdd_AddCache(BddRelOpComplexGarbage);

}

/**Function****************************************************************
  Synopsis    [Function Bdd_Exit]
  Description [Bdd_ExitPkg deallocates all memory allocated by Bdd_InitPkg
               Bdd_FoaNode, Bdd_FoaTerminal ets.]
  SideEffects [Currently, function does not deallocate all memory!]
  SeeAlso     [Bdd_InitPkg]
  ************************************************************************/

void
Bdd_ExitPkg()
{
  int i;
  Bdd_Node *sup, *sup1;

  if (bdd_status == 0) {
    printf("\nBDD package is not initialized. ");
    return;
  } else {
    bdd_status = 0;
  }

  /* DELETE NODES */
  for (i=0; i<bddNodeTable.size; i++) {
    sup = bddNodeTable.table[i];
    if (sup) sup = sup->next; /* do not delete first element! */
    while (sup) {
      sup1 = sup->next;
      free(sup);
      sup = sup1;
    }
  }

  /* DELETE NODES, WHICH ARE FIRST ELEMENT IN NODE TABLE */
  free(bddNodeTable.table);

  /* DELETE SYMBOL TREE */
  /* not implemented */

  /* DELETE FORMULAE TREE */
  /* not implemented */

  /* DELETE ITE CACHE */
  free(bddIteCache.table);

  /* DELETE EAX CACHE TABLE */
  free(bddEAxCache.table);

  /* DELETE RelOp CACHE TABLE */
  free(bddRelOpCache.table);
}

/**Function****************************************************************
  Synopsis    [Function Bdd_AboutPkg]
  Description [Bdd_AboutPkg reports release of Binary Decision
               Diagrams package.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_AboutPkg() {
  printf("Binary Decision Diagrams package, release ");
  printf(BDD_RELEASE);
  printf("\n");
}

/**Function****************************************************************
  Synopsis    [Function Bdd_FindTerminal]
  Description [Bdd_FindTerminal find a terminal node.
               Terminal node is a node with given variable and successors
               bdd_termFalse and bdd_termTrue. If such terminal node
               does not exists, function return bdd_termNull.]
  SideEffects []
  SeeAlso     [Bdd_FoaTerminal]
  ************************************************************************/

Bdd_Edge
Bdd_FindTerminal(Bdd_String x)
{
  Bdd_Variable *sup;
  Bdd_Boolean end, find;

  /* SYMBOL TREE IS NEVER EMPTY. AT LEAST, THERE IS ELEMENT '1' */
  sup = bddSymbolTree;
  end = FALSE;
  find = FALSE;

  /* LOOKING FOR AN ELEMENT */
  while (!(find || end)) {
    if (strcmp(x, sup->name) < 0) {
      if (sup->l) sup = sup->l;
	else end = TRUE;
    } else {
      if (strcmp(x, sup->name) > 0) {
	if (sup->r) sup = sup->r;
	  else end = TRUE;
      } else {
	find = TRUE;
      }
    }
  }

  /* IF NOT FOUND... */
  if (!find) return bdd_termNull;

  /* SUP IS NOW THE ELEMENT, WHICH HAS BEEN LOOKED FOR */
  return sup->term;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_FoaTerminal]
  Description [Bdd_FoaTerminal find or add terminal node.
               Terminal node is a node with given variable and successors
               bdd_termFalse and bdd_termTrue. If such terminal node already
               exists, function return it and does not create the new one.]
  SideEffects []
  SeeAlso     [Bdd_FindTerminal, Bdd_FoaNode]
  ************************************************************************/

Bdd_Edge
Bdd_FoaTerminal(Bdd_String x)
{
  Bdd_Variable *sup;
  Bdd_Boolean end, find;
  int n;

  /* SYMBOL TREE IS NEVER EMPTY. AT LEAST, THERE IS ELEMENT '1' */
  sup = bddSymbolTree;
  end = FALSE; 
  find = FALSE;

  /* LOOKING FOR THE ELEMENT */
  while (!(find || end)) {
    if (strcmp(x, sup->name) < 0) {
      if (sup->l) sup = sup->l;
        else end = TRUE;
    } else {
      if (strcmp(x, sup->name) > 0) {
	if (sup->r) sup = sup->r;
	  else end = TRUE;
      } else {
	find = TRUE;
      }
    }
  }

  /* IF THE ELEMENT WAS NOT FOUND... */
  if (!find) {

    if (strcmp(x, sup->name) < 0) {
      if (!(sup->l = (Bdd_Variable *) malloc(sizeof(Bdd_Variable)))) {
        fprintf(stdout,"Bdd_FoaTerminal: Out of memory!\n");
        exit(1);
      }
      sup->l->parent = sup;
      sup = sup->l;
    } else {
      if (!(sup->r = (Bdd_Variable *) malloc(sizeof(Bdd_Variable)))) {
        fprintf(stdout,"Bdd_FoaTerminal: Out of memory!\n");
        exit(1);
      }
      sup->r->parent = sup;
      sup = sup->r;
    }

    /* SYMBOL TREE COULD BE A RED-BLACK TREE */
    /* http://mathworld.wolfram.com/Red-BlackTree.html */
    /* 1. Every node has two children, each colored either red or black. */
    /* 2. Every tree leaf node is colored black. */
    /* 3. Every red node has both of its children colored black. */
    /* 4. Every path from the root contains the same number of black nodes. */

    /* ALGORITHMS FOR BALANCING RED-BLACK TREES ARE NOT IMPLEMENTED, YET */

    /* sup IS NOW THE CREATED ELEMENT */

    sup->l = NULL;
    sup->r = NULL;
    sup->name = strdup(x);
    sup->value = FALSE;
    n = bdd_termTrue.p->v->order;
    sup->order = n;
    bdd_termTrue.p->v->order = n+1;
    sup->term = Bdd_FoaNode(sup, bdd_termFalse, bdd_termTrue);
    sup->term.p->count = 0;
    bddNodeTable.nf--;
    Bdd_AddFormula(x,sup->term);
  }

  /* sup IS NOW THE ELEMENT, WHICH HAS BEEN LOOKED FOR */
  return sup->term;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_FoaNode]
  Description [Bdd_FoaNode find or add new node with given variable
               and successors. If such node already exists, function
               return it and does not create the new one.]
  SideEffects [BE CAREFULL WITH TYPE Bdd_Node!]
  SeeAlso     [Bdd_FoaTerminal]
  ************************************************************************/

Bdd_Edge
Bdd_FoaNode(Bdd_Variable *v, Bdd_Edge pf, Bdd_Edge pt)
{
  unsigned int hash;
  Bdd_Edge edge;
  Bdd_Node *sup, *sup1;
  int i;

  if ((pf.p->count) && (pf.p->count != bdd_counter)) {
    fprintf(stdout,
            "ERROR (Bdd_FoaNode): Bad pf (bdd_counter = %d, count = %d)!\n",
            bdd_counter,(pf.p)->count);
    sleep(5);
    exit(1);
  }

  if ((pt.p->count) && (pt.p->count != bdd_counter)) {
    fprintf(stdout,
            "ERROR (Bdd_FoaNode): Bad pt (bdd_counter = %d, count = %d)!\n",
            bdd_counter,(pt.p)->count);
    sleep(5);
    exit(1);
  }

  bddNodeTable.foa++;

  /* MINIMIZATION */

  if (Bdd_Cmp(pf,pt)) return pf;

  /* NORMALIZATION - EDGE 'THEN' MAY NOT BE COMPLEMENTED ! */
  if (pt.mark) {
    edge.mark = TRUE;
    pf.mark = !pf.mark;
    pt.mark = FALSE;
  } else {
    edge.mark = FALSE;
  }

  /* THIS STATEMENT ENABLES GARBAGE COLLECTION */
  /* IF THERE IS SOMETHING TO DELETE AND IF ALL GENERATED */
  /* NODES ARE USED THEN TRY GARBAGE COLLECTION           */
  /**/
  if (bddNodeTable.num >= bddNodeTable.generated && 
        bddFirstNewNode != bddFreshNodes) Bdd_Garbage();
  /**/

  /* THIS IS HASH FUNCTION FOR NODE TABLE */
  /* BECAUSE OF USED TRICKS, HASH FUNCTION MUST NOT RETURN ZERO! */

  if (!(hash = (((unsigned int) v->order + (unsigned int) pf.p +
		    (unsigned int) pt.p) % bddNodeTable.size))) hash=1;

  /* FIND OR ADD - THERE IS A HASH TABLE WITH CHAINING */

  sup = bddNodeTable.table[hash];
  sup1 = NULL;

  /* ********************************************************************* */
  /* USE THIS IF YOU WANT ORDERED CHAINS IN HASH TABLE                     */
  /* THIS MUST BE COMBINED WITH ADDING NODES AFTER sup1                    */
  /**/

  while (sup && (v->order < sup->v->order))
  {
    bddNodeTable.compare++;
    sup1 = sup;
    sup = sup->next;
  }

  while (sup && (v->order == sup->v->order) &&
           ((pf.p != sup->f.p) || (pt.p != sup->t.p) ||
            (pf.mark != sup->f.mark) || (pt.mark != sup->t.mark)))
  {
    bddNodeTable.compare++;
    sup1 = sup;
    sup = sup->next;
  }
  /**/
  /* ********************************************************************* */


  /* ********************************************************************* */
  /* USE THIS FOR ADDING ON THE BEGINNING OR ON THE END OF CHAIN           */
  /* ********************************************************************* */
  /*

  while (sup &&
           ((pf.p != sup->f.p) || (pt.p != sup->t.p) ||
            (pf.mark != sup->f.mark) || (pt.mark != sup->t.mark) || (v != sup->v)))
  {
    bddNodeTable.compare++;
    sup1 = sup;
    sup = sup->next;
  }
  */
  /* ********************************************************************* */

  if ((!sup) || (v != sup->v)) { 

    /* NEW NODE MUST BE ADDED */

    /* ONE FRESH NODE WILL BE ADDED */
    bddNodeTable.num++;
    bddNodeTable.add++;
    bddNodeTable.nf++;
    if (bddNodeTable.num > bddNodeTable.max)
          bddNodeTable.max = bddNodeTable.num;

    if (!bddFreeNodes) {

      /* THESE LINES DISABLE ALLOCATION OF MEMORY AFTER INITIALIZATION */
      /*
      fprintf(stdout,"\nBDD (Bdd_FoaNode): Allocated node table is full!\n");
      fprintf(stdout,"Currently, there exist %d nodes.\n",bddNodeTable.num);
      Bdd_SystemStat();
      exit(1);
      */

      if (!(bddFreeNodes = (Bdd_Node *)
                           malloc((bdd_maxNodes) * sizeof(Bdd_Node)))) {
        fprintf(stdout,"\nBDD (Bdd_FoaNode): Out of memory error!\n");
	fprintf(stdout,"Currently, there exist %d nodes.\n",bddNodeTable.num);
        Bdd_SystemStat();
        exit(1);
      }

      for (i=0; i<bdd_maxNodes-1; i++)
        bddFreeNodes[i].new = &bddFreeNodes[i+1];
      bddFreeNodes[bdd_maxNodes-1].new = NULL;

      bddNodeTable.generated = bddNodeTable.generated + bdd_maxNodes;
    }

    sup = bddFreeNodes;
    bddFreeNodes = sup->new;

    /* BE CAREFULLY WITH TYPE Bdd_Node !!!! */
    /* THIS COMPUTATION USE SOME TRICKS! */
    /* FIELDS 'prev' AND 'next' HAVE TO BE THE FIRST AND THE SECOND! */

    /* ADD NODE ON THE FIRST PLACE IN THE CHAIN */
    /*
    sup->prev = (Bdd_Node *) &bddNodeTable.table[hash-1];
    sup->next = bddNodeTable.table[hash];
    bddNodeTable.table[hash] = sup;
    if (sup->next) sup->next->prev = sup;
    */

    /* ADD NODE AFTER sup1 */
    /**/
    sup->prev = sup1 ? sup1 : (Bdd_Node *) &bddNodeTable.table[hash-1];
    sup->next = sup->prev->next;
    sup->prev->next = sup;
    if (sup->next) sup->next->prev = sup;
    /**/

    bddNewNodes->new = sup;
    bddNewNodes = sup;
    sup->f = pf;
    sup->t = pt;
    sup->v = v;
    sup->count = bdd_counter;

  } else {
    if (sup->count) Bdd_Fresh(sup);
  }
  edge.p = sup;
  return edge;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_NOT]
  Description [Bdd_NOT calculates negation of Boolean function.]
  SideEffects []
  SeeAlso     [Bdd_ITE, Bdd_Expr]
  ************************************************************************/

Bdd_Edge
Bdd_NOT(Bdd_Edge f)
{
  f.mark = !f.mark;
  return f;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_ITE]
  Description [Bdd_ITE calculates ITE operation of three Boolean functions.
               If an error occurs, please, report it to the author.]
  SideEffects []
  SeeAlso     [Bdd_NOT, Bdd_Expr]
  ************************************************************************/

Bdd_Edge
Bdd_ITE(Bdd_Edge f, Bdd_Edge g, Bdd_Edge h)
{
  Bdd_Boolean negation;
  Bdd_Edge r, T, E, Fv, Gv, Hv, Fneg_v, Gneg_v, Hneg_v;
  Bdd_Variable *v;

  static Bdd_Boolean terminal; /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_Variable *topF;   /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_Variable *topG;   /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_Variable *topH;   /* STATIC TO MINIMIZE STACK PROBLEMS */

    if (!f.p) {
      fprintf(stdout,"ERROR (Bdd_ITE): f = NULL!\n");
      sleep(5);
      exit(1);
    }

    if (!g.p) {
      fprintf(stdout,"ERROR (Bdd_ITE): g = NULL!\n");
      sleep(5);
      exit(1);
    }

    if (!h.p) {
      fprintf(stdout,"ERROR (Bdd_ITE): h = NULL!\n");
      sleep(5);
      exit(1);
    }

    if ((f.p->count) && (f.p->count != bdd_counter)) {
      fprintf(stdout,
              "ERROR (Bdd_ITE): Bad f (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(f.p)->count);
      sleep(5);
      exit(1);
    }

    if ((g.p->count) && (g.p->count != bdd_counter)) {
      fprintf(stdout,
             "ERROR (Bdd_ITE): Bad g (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(g.p)->count);
      sleep(5);
      exit(1);
    }

    if ( (h.p->count) && (h.p->count != bdd_counter) ) {
      fprintf(stdout,
              "ERROR (Bdd_ITE): Bad h (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(h.p)->count);
      sleep(5);
      exit(1);
    }

    /* NORMALIZATION - STEP 1 */
    if (f.p == g.p) {
      g.p = bdd_termTrue.p;
      g.mark = (f.mark != g.mark);
    }
    
    if (f.p == h.p) {
      h.p = bdd_termTrue.p;
      h.mark = (f.mark == h.mark);
    }

    /* NORMALIZATION - STEP 2 */
    if (g.p == bdd_termTrue.p) {
      if ((unsigned int) f.p > (unsigned int) h.p) {
	if (g.mark) {
	  Bdd_ExchangeNeg(&f, &h);
	} else {
	  Bdd_Exchange(&f, &h);
        }
      }
    }

    if (h.p == bdd_termTrue.p) {
      if ((unsigned int) f.p > (unsigned int) g.p) {
	if (h.mark) {
	 Bdd_Exchange(&f, &g);
	} else {
	  Bdd_ExchangeNeg(&f, &g);
        }
      }
    }

    if (g.p == h.p) {
      if (g.mark != h.mark) {
        if ((unsigned int) f.p > (unsigned int) g.p) {
          h.p = f.p;
	  h.mark = ! f.mark;
	  Bdd_Exchange(&f, &g);
	}
      }
    }

    negation = FALSE;
  
    /* LOOKING FOR TERMINAL CASE */

    terminal = FALSE;
    if (f.p == bdd_termTrue.p) {
      terminal = TRUE;
      r = f.mark ? h : g;
    } else {
      if (g.p == h.p) {
        if (g.mark == h.mark) {
	  terminal = TRUE;
	  r = g;
	} else {
          if (g.p == bdd_termTrue.p) {
	    terminal = TRUE;
	    r = f;
	    negation = g.mark;
	  }
        }
      }
    }

    /* IF NOT TERMINAL CASE ==> NORMALIZATION - STEP 3 */
    if (!terminal) {

      if (f.mark) {
        if (h.mark) {
          f.mark = FALSE;
          negation = TRUE;
          Bdd_ExchangeNeg(&g, &h);
	} else {
          f.mark = FALSE;
          Bdd_Exchange(&g, &h);
	}
      } else {
	if (g.mark) {
          negation = TRUE;
          g.mark = FALSE;
          h.mark = ! h.mark;
	}
      }

      /* IF RESULT IS NOT IN THE CACHE TABLE... */
      if (!BddFindIteCache(f, g, h, &r)) {

         /* LOOKING FOR THE SMALLES TOP VARIABLE */
         topF = (f.p)->v;
         topG = (g.p)->v;
         topH = (h.p)->v;

         if (topF->order < topG->order || g.p == bdd_termTrue.p) {
           v = ((topF->order < topH->order) || (h.p == bdd_termTrue.p)) ?
                 topF : topH;
         } else {
           v = ((topH->order < topG->order) && (h.p != bdd_termTrue.p)) ?
                 topH : topG;
         }

         /* DETERMINING PARAMETERS FOR RECURSIVE CALLS */
         if (topF == v) {
           if (f.mark) {
             Fv = Bdd_NOT(f.p->t);
             Fneg_v = Bdd_NOT(f.p->f);
           } else {
             Fv = f.p->t;
             Fneg_v = f.p->f;
           }
         } else {
           Fneg_v = Fv = f;
         }

         if (topG == v) {
           if (g.mark) {
             Gv = Bdd_NOT(g.p->t);
             Gneg_v = Bdd_NOT(g.p->f);
           } else {
             Gv = g.p->t;
             Gneg_v = g.p->f;
           }
         } else {
           Gneg_v = Gv = g;
         }

         if (topH == v) {
           if (h.mark) {
             Hv = Bdd_NOT(h.p->t);
             Hneg_v = Bdd_NOT(h.p->f);
           } else {
             Hv = h.p->t;
             Hneg_v = h.p->f;
           }
         } else {
           Hneg_v = Hv = h;
         }

         /* REKURSIVE CALLS */
         T = Bdd_ITE(Fv, Gv, Hv);
         E = Bdd_ITE(Fneg_v, Gneg_v, Hneg_v);

         r = Bdd_FoaNode(v, E, T);
         BddAddIteCache(f, g, h, r);

      } else {

         /* IF THE RESULT IS FROM CACHE TABLE, FRESH IT! */
         Bdd_Fresh(r.p);
      }
    }

    if (negation) r.mark = ! r.mark;
    return r;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Restrict]
  Description [Bdd_Restrict calculates a restriction of Boolean function.
               Function does not use hash table.]
  SideEffects []
  SeeAlso     [Bdd_Compose, Bdd_E, Bdd_A]
  ************************************************************************/

Bdd_Edge
Bdd_Restrict(Bdd_Edge f, Bdd_Variable *v, Bdd_Boolean value)
{
  Bdd_Edge l, d, r;

  if ( (f.p->count) && (f.p->count != bdd_counter) ) {
    fprintf(stdout,
            "ERROR (Bdd_Restrict): Bad f (bdd_counter = %d, count = %d)!\n",
            bdd_counter,f.p->count);
    sleep(5);
    exit(1);
  }

  if (f.p == bdd_termTrue.p) return f;
  if (f.p->v->order > v->order) return f;
  if (f.p->v == v) {
    if (value) return Bdd_TransferMark(f.p->t, f.mark);
      else return Bdd_TransferMark(f.p->f, f.mark);
  }

  l = Bdd_Restrict(f.p->f, v, value);
  d = Bdd_Restrict(f.p->t, v, value);
  r =  Bdd_TransferMark(Bdd_FoaNode(f.p->v, l, d), f.mark);
  return r;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Compose]
  Description [Bdd_Compose calculates a composition of two Boolean
               functions. Function does not use hash table.]
  SideEffects []
  SeeAlso     [Bdd_Restrict, Bdd_E, Bdd_A]
  ************************************************************************/

Bdd_Edge
Bdd_Compose(Bdd_Edge f, Bdd_Variable *v, Bdd_Edge g)
{
  Bdd_Edge l, d, r;

  if ( (f.p->count) && (f.p->count != bdd_counter) ) {
    fprintf(stdout,
            "ERROR (Bdd_Compose): Bad f (bdd_counter = %d, count = %d)!\n",
            bdd_counter,f.p->count);
    sleep(5);
    exit(1);
  }

  if (f.p == bdd_termTrue.p) return f;
  if (f.p->v->order > v->order) return f;
  if (f.p->v == v) {
    return Bdd_TransferMark(Bdd_ITE(g,f.p->t,f.p->f), f.mark);
  }

  l = Bdd_Compose(f.p->f, v, g);
  d = Bdd_Compose(f.p->t, v, g);
  r =  Bdd_TransferMark(Bdd_FoaNode(f.p->v, l, d), f.mark);
  return r;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_E]
  Description [Bdd_E calculates an existential quantification of
               Boolean function. Function does not use hash table.]
  SideEffects []
  SeeAlso     [Bdd_Restrict, Bdd_Compose, Bdd_A]
  ************************************************************************/

Bdd_Edge
Bdd_E(Bdd_Edge f, Bdd_Variable *v)
{
  Bdd_Edge l, d, r;

  if ( (f.p->count) && (f.p->count != bdd_counter) ) {
    fprintf(stdout,
            "ERROR (Bdd_E): Bad f (bdd_counter = %d, count = %d)!\n",
            bdd_counter,f.p->count);
    sleep(5);
    exit(1);
  }

  if (f.p == bdd_termTrue.p) return f;
  if (f.p->v->order > v->order) return f;
  if (f.p->v == v) {
    return Bdd_ITE(Bdd_TransferMark(f.p->t, f.mark),
                   bdd_termTrue,
                   Bdd_TransferMark(f.p->f, f.mark));
  }

  l = Bdd_E(f.p->f, v);
  d = Bdd_E(f.p->t, v);
  r =  Bdd_TransferMark(Bdd_FoaNode(f.p->v, l, d), f.mark);
  return r;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_A]
  Description [Bdd_A calculates an universal quantification of
               Boolean function. Function does not use hash table.]
  SideEffects []
  SeeAlso     [Bdd_Restrict, Bdd_Compose, Bdd_A]
  ************************************************************************/

Bdd_Edge
Bdd_A(Bdd_Edge f, Bdd_Variable *v)
{
  Bdd_Edge l, d, r;

  if ( (f.p->count) && (f.p->count != bdd_counter) ) {
    fprintf(stdout,
            "ERROR (Bdd_A): Bad f (bdd_counter = %d, count = %d)!\n",
            bdd_counter,f.p->count);
    sleep(5);
    exit(1);
  }

  if (f.p == bdd_termTrue.p) return f;
  if (f.p->v->order > v->order) return f;
  if (f.p->v == v) {
    return Bdd_ITE(Bdd_TransferMark(f.p->t, f.mark),
                   Bdd_TransferMark(f.p->f, f.mark),
                   bdd_termFalse);
  }

  l = Bdd_A(f.p->f, v);
  d = Bdd_A(f.p->t, v);
  r =  Bdd_TransferMark(Bdd_FoaNode(f.p->v, l, d), f.mark);
  return r;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_IncCounter]
  Description [Bdd_IncCounter increment variable bdd_counter by 1.]
  SideEffects [bdd_counter may not exceed MaxInt (2147483647)]
  SeeAlso     [Garbage Collection]
  ************************************************************************/

void
Bdd_IncCounter()
{
  bdd_counter++;
  bddFreshNodes = bddNewNodes;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Fresh]
  Description [Bdd_Fresh fresh all non-fortified nodes in BDD]
  SideEffects []
  SeeAlso     [Garbage Collection]
  ************************************************************************/

void
Bdd_Fresh(Bdd_Node *f)
{
  if (f->count && f->count != bdd_counter) {
    f->count = bdd_counter;
    Bdd_Fresh(f->f.p);
    Bdd_Fresh(f->t.p);
  }
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Fortify]
  Description [Bdd_Fresh fortify all nodes in BDD]
  SideEffects []
  SeeAlso     [Garbage Collection]
  ************************************************************************/

void
Bdd_Fortify(Bdd_Node *f)
{
  if (f->count) {
    f->count = 0;
    bddNodeTable.nf--;
    Bdd_Fortify(f->f.p);
    Bdd_Fortify(f->t.p);
  }
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Garbage]
  Description [Bdd_Garbage perform garbage collection.]
  SideEffects []
  SeeAlso     [Garbage Collection]
  ************************************************************************/

void
Bdd_Garbage()
{
  Bdd_Node *sup;
  BddCache *c;
  static int n = 0;

  /*
  fprintf(stdout,"Garbage (%d)...\n",++n);
  */

  /* Updating cache tables */
  if (bddFirstNewNode != bddFreshNodes) {
    bddNodeTable.garbage++;

    c = bddCacheList;
    while (c) {
      c->gc();
      c = c->next;
    }

    /* BE CAREFULLY WITH TYPE Bdd_Node !!!! */
    /* THIS COMPUTATION USE SOME TRICKS! */
    /* FIELDS 'prev' AND 'next' HAVE TO BE THE FIRST AND THE SECOND! */

    /* Updating unique table */
    while (bddFirstNewNode != bddFreshNodes) {
      if (!bddFirstNewNode->count) {
        bddFirstNewNode = bddFirstNewNode->new;
      } else {
        if (bddFirstNewNode->count == bdd_counter) {
          bddNewNodes->new = bddFirstNewNode;
          bddNewNodes = bddFirstNewNode;
          bddFirstNewNode = bddFirstNewNode->new;
        } else {
          bddFirstNewNode->prev->next = bddFirstNewNode->next;
          if (bddFirstNewNode->next)
            bddFirstNewNode->next->prev = bddFirstNewNode->prev;
          sup = bddFirstNewNode->new;
          bddFirstNewNode->new = bddFreeNodes;
          bddFreeNodes = bddFirstNewNode;
          bddFirstNewNode = sup;
          bddNodeTable.num--;
          bddNodeTable.nf--;
        }
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function Bdd_TransferMark]
  Description [Bdd_TransferMark complements Boolean function if
               given parameter mark is TRUE.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_TransferMark(Bdd_Edge f, Bdd_Boolean mark)
{
  f.mark = (f.mark != mark);
  return f;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_Exchange]
  Description [Bdd_Exchange exchange two Boolean functions.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void Bdd_Exchange(Bdd_Edge *f, Bdd_Edge *g)
{
  Bdd_Edge sup;

  sup = *f;
  *f = *g;
  *g = sup;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_ExchangeNeg]
  Description [Bdd_ExchangeNeg complements and exchange two Boolean
               functions.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void Bdd_ExchangeNeg(Bdd_Edge *f, Bdd_Edge *g)
{
  Bdd_Edge sup;

  sup = *f;
  *f = *g;
  *g = sup;
  (*f).mark = !(*f).mark;
  (*g).mark = !(*g).mark;
}


/**Function****************************************************************
  Synopsis    [Function Bdd_FindByOrder]
  Description [Bdd_FindByOrder finds variable with the given order.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Variable
*Bdd_FindByOrder(Bdd_Variable *v, int n)
{
  Bdd_Variable *sup;

  if (v == NULL) return NULL;
  if (v->order == n) return v;
  sup = Bdd_FindByOrder(v->l,n);
  if (sup != NULL) return sup;
  return(Bdd_FindByOrder(v->r,n));
}

/**Function****************************************************************
  Synopsis    [Function Bdd_FindFormula]
  Description [Bdd_FindFormula find formula in Fomulea tree.]
  SideEffects []
  SeeAlso     [Bdd_AddFormula]
  ************************************************************************/

Bdd_Boolean
Bdd_FindFormula(Bdd_String x, Bdd_Edge *f)
{
  BddFormula *sup;

  sup = bddFormulaeTree;
  while (sup && strcmp(sup->name, x)) {
    sup = strcmp(x, sup->name) < 0 ? sup->l : sup->r;
  }
  if (!sup) return FALSE;
  *f = sup->p;
  return TRUE;
}

/**Function****************************************************************
  Synopsis    [Function Bdd_AddFormula]
  Description [Bdd_AddFormula add formula to the Formulae tree.
               If such formula already exists, function overwrite it!]
  SideEffects [If bddFormulaeTree does not exist, function create it]
  SeeAlso     [Bdd_FindFormula]
  ************************************************************************/

void
Bdd_AddFormula(Bdd_String x, Bdd_Edge f)
{
  BddFormula *sup;
  Bdd_Boolean end, overwrite;

  overwrite = FALSE;
  if (!bddFormulaeTree) {

    if (!(bddFormulaeTree = (BddFormula *) malloc(sizeof(BddFormula)))) {
      fprintf(stdout,"Bdd_AddFormula: Out of memoy!\n");
      exit(1);
    }

    bddFormulaeTree->l = bddFormulaeTree->r = NULL;
    bddFormulaeTree->name = strdup(x);
    bddFormulaeTree->p = f;
  } else {
    end = FALSE;
    sup = bddFormulaeTree;
    while (!end) {
      if (strcmp(x, sup->name) < 0) {
	if (sup->l) sup = sup->l;
	  else end = TRUE;
      } else {
        if (strcmp(x, sup->name) > 0) {
	  if (sup->r) sup = sup->r;
	    else end = TRUE;
        } else {
          sup->p = f;
          overwrite = TRUE;
          end = TRUE;
        }
      }
    }

    if (!overwrite) {
      if (strcmp(x, sup->name) < 0) {

        if (!(sup->l = (BddFormula *) malloc(sizeof(BddFormula)))) {
          fprintf(stdout,"Bdd_AddFormula: Out of memoy!\n");
          exit(1);
        }

	sup = sup->l;

      } else {

        if (!(sup->r = (BddFormula *) malloc(sizeof(BddFormula)))) {
          fprintf(stdout,"Bdd_AddFormula: Out of memoy!\n");
          exit(1);
        }

	sup = sup->r;
      }
      sup->l = sup->r = NULL;
      sup->name = strdup(x);
      sup->p = f;
    }
  }
}


/**Function****************************************************************
  Synopsis    [Function Bdd_SaveFormula]
  Description [Bdd_SaveFormula add function to the Formulae tree.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
Bdd_SaveFormula(Bdd_String prefix, Bdd_String name, Bdd_Edge f)
{
  Bdd_String s;

  s = strdup(prefix);
  s = (char*) realloc(s, strlen(prefix) + strlen(name) + 1);
  strcat(s,name);
  Bdd_AddFormula(s,f);
  free(s);
}

/**Function****************************************************************
  Synopsis    [Function Bdd_AddCache]
  Description [Bdd_AddCache add cache to the end of Cache list]
  SideEffects [If bddCacheList does not exist, function create it]
  SeeAlso     [Garbage Collection]
  ************************************************************************/

void
Bdd_AddCache(Bdd_UserFunction gc)
{
  BddCache *sup;

  if (!bddCacheList) {

    if (!(bddCacheList = (BddCache *) malloc(sizeof(BddCache)))) {
      fprintf(stdout,"Bdd_AddCache: Out of memoy!\n");
      exit(1);
    }

    bddCacheList->gc = gc;
    bddCacheList->next = NULL;
  } else {
    sup = bddCacheList;
    while (sup->next) sup = sup->next;

    if (!(sup->next = (BddCache *) malloc(sizeof(BddCache)))) {
      fprintf(stdout,"Bdd_AddCache: Out of memoy!\n");
      exit(1);
    }

    sup = sup->next;
    sup->gc = gc;
    sup->next = NULL;
  }
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function BddAddIteCache]
  Description [BddAddIteCache add parameters and result to the ITE cache.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddAddIteCache(Bdd_Edge a, Bdd_Edge b, Bdd_Edge c, Bdd_Edge r)
{
  int indeks;
  BddIteCache *p;

  /* THIS IS HASH FUNCTION */
  /* MUST BE THE SAME AS IN BddFindIteCache! */

  indeks = (int) ((
                    ((unsigned int) a.p << 16) +
                    ((unsigned int) b.p << 8) +
                    (unsigned int) c.p
                  ) % bddIteCache.size);

  p = &bddIteCache.table[indeks];
  if (!p->ok)
    {
      p->ok = TRUE;
      p->f = a.p;
      p->g = b.p;
      p->h = c.p;
      p->hmark = c.mark;
      p->result = r;
    }

#ifdef MINACC
  else
    if (p->f->count || p->g->count || p->h->count || p->result.p->count)
      {
        bddIteCache.overwrite++;
        p->f = a.p;
        p->g = b.p;
        p->h = c.p;
        p->hmark = c.mark;
        p->result = r;
      }
#endif

#ifdef MAXACC
  else
      {
	bddIteCache.overwrite++;
        p->f = a.p;
        p->g = b.p;
        p->h = c.p;
        p->hmark = c.mark;
        p->result = r;
      }
#endif

}

/**Function****************************************************************
  Synopsis    [Function BddFindIteCache]
  Description [BddFindIteCache looks for result in the ITE cache.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Boolean
BddFindIteCache(Bdd_Edge a, Bdd_Edge b, Bdd_Edge c, Bdd_Edge *r)
{
  int indeks;
  BddIteCache *p;

  /* THIS IS HASH FUNCTION */
  /* MUST BE THE SAME AS IN BddAddIteCache! */

  indeks = (int) ((
                    ((unsigned int) a.p << 16) +
                    ((unsigned int) b.p << 8) +
                    (unsigned int) c.p
                  ) % bddIteCache.size);

  bddIteCache.search++;
  p = &bddIteCache.table[indeks];

  if (p->ok && p->f == a.p && p->g == b.p &&
      p->h == c.p && p->hmark == c.mark)
  {
    bddIteCache.find++;
    *r = p->result;
    return TRUE;
  }
  return FALSE;
}

/**Function****************************************************************
  Synopsis    [Function BddIteGarbage.]
  Description [BddIteGarbage performs Garbage Collection for ITE cache]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddIteGarbage()
{
  int j;
  BddIteCache *c;

  for (j=0; j<bddIteCache.size; j++) {
    if (bddIteCache.table[j].ok) {
      c = &bddIteCache.table[j];
      if ((c->f->count && c->f->count != bdd_counter) ||
          (c->g->count && c->g->count != bdd_counter) ||
          (c->h->count && c->h->count != bdd_counter) ||
	  (c->result.p->count && c->result.p->count != bdd_counter))
      {
        c->ok = FALSE;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function BddEAxGarbage.]
  Description [BddEAxGarbage performs Garbage Collection for EAx cache]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddEAxGarbage()
{
  int j;
  BddEAxCache *c;

  for (j=0; j<bddEAxCache.size; j++) {
    if (bddEAxCache.table[j].ok) {
      c = &bddEAxCache.table[j];
      if ((c->f->count && c->f->count != bdd_counter) ||
	  (c->result.p->count && c->result.p->count != bdd_counter))
      {
        c->ok = FALSE;
      }
    }
  }
}
