/**CFile*******************************************************************
  PackageName [bdd]
  FileName    [bddRel.c]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]

  Revision    [$Revision: 2.4 $]
  Date        [$Date: 2005/01/25 10:06:23 $]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]
  Description [The file bddRel.c contains functions for relational
               operations (e.g. RelProd).]
  SeeAlso     [bdd.h, bddInt.h]
  Copyright   [Copyright (C) 2000-2004 Robert Meolic
               This program is free software; you can redistribute it
               and/or modify it under the terms of the GNU General Public
               License (version 2) as published by the Free Software
               Foundation.

               This program is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 59 Temple Place, Suite 330,
               Boston, MA  02111-1307 USA.]
  ************************************************************************/

#include "bddInt.h"

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

#define oNULL 0
#define oAND 1
#define oOR 2
#define oIMPL 3

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

BddRelOpCacheTable bddRelOpCache;              /* Operations cache table */

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_RelOp(Bdd_Edge f, Bdd_Edge g, Bdd_String format, Bdd_Boolean firsttime)
{
  static int op;
  static unsigned int key;

  BddRelOpCache *p;
  Bdd_Variable *z;
  Bdd_Edge f0, f1, g0, g1, h0, h1, h;

  static Bdd_String v;     /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_Boolean done; /* STATIC TO MINIMIZE STACK PROBLEMS */

  Bdd_Boolean Ex;
  Bdd_Boolean Ax;
  static Bdd_Boolean ExA;
  static Bdd_Boolean AxA;
  static Bdd_Boolean ExB;
  static Bdd_Boolean AxB;
  static Bdd_Boolean ExR;
  static Bdd_Boolean AxR;
  static Bdd_Boolean ExS;
  static Bdd_Boolean AxS;
  static Bdd_Boolean ExP;
  static Bdd_Boolean AxP;
  static Bdd_Boolean ExQ;
  static Bdd_Boolean AxQ;

  static Bdd_Boolean A2B;
  static Bdd_Boolean R2S;
  static Bdd_Boolean R2P;
  static Bdd_Boolean R2Q;
  static Bdd_Boolean S2R;
  static Bdd_Boolean S2P;
  static Bdd_Boolean S2Q;
  static Bdd_Boolean P2Q;
  static Bdd_Boolean P2R;
  static Bdd_Boolean Q2P;

  /* IF FIRST-TIME, PARSING FORMAT STRING */

  if (firsttime) {

    if ((f.p->count) && (f.p->count != bdd_counter)) {
      fprintf(stderr,
              "ERROR (Bdd_RelOp): Bad f (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(f.p)->count);
      sleep(5);
      exit(1);
    }

    if ((g.p->count) && (g.p->count != bdd_counter)) {
      fprintf(stderr,
             "ERROR (Bdd_RelOp): Bad g (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(g.p)->count);
      sleep(5);
      exit(1);
    }

    key = 0;                               /* 0 * 2^30 */
    op = oNULL;

    if (strstr(format,"#AND")) {
      key = 1073741824U;                   /* 1 * 2^30 */
      op = oAND;
    }

    if (strstr(format,"#OR")) {
      key = 2147483648U;                   /* 2 * 2^30 */
      op = oOR;
    }

    if (strstr(format,"#IMPL")) {
      key = 3221225472U;                   /* 3 * 2^30 */
      op = oIMPL;
    }

    Ex = (strstr(format,"Ex") != NULL);
    Ax = (strstr(format,"Ax") != NULL);

    if (Ex) {
      ExA = (strstr(format,"xA") != NULL);
      ExB = (strstr(format,"xB") != NULL);
      ExR = (strstr(format,"xR") != NULL);
      ExS = (strstr(format,"xS") != NULL);
      ExP = (strstr(format,"xP") != NULL);
      ExQ = (strstr(format,"xQ") != NULL);
    } else {
      ExA = FALSE;
      ExB = FALSE;
      ExR = FALSE;
      ExS = FALSE;
      ExP = FALSE;
      ExQ = FALSE;
    }

    if (Ax) {
      AxA = (strstr(format,"xA") != NULL);
      AxB = (strstr(format,"xB") != NULL);
      AxR = (strstr(format,"xR") != NULL);
      AxS = (strstr(format,"xS") != NULL);
      AxP = (strstr(format,"xP") != NULL);
      AxQ = (strstr(format,"xQ") != NULL);
    } else {
      AxA = FALSE;
      AxB = FALSE;
      AxR = FALSE;
      AxS = FALSE;
      AxP = FALSE;
      AxQ = FALSE;
    }

    A2B = (strstr(format,"A2B") != NULL);
    R2S = (strstr(format,"R2S") != NULL);
    R2P = (strstr(format,"R2P") != NULL);
    R2Q = (strstr(format,"R2Q") != NULL);
    S2R = (strstr(format,"S2R") != NULL);
    S2P = (strstr(format,"S2P") != NULL);
    S2Q = (strstr(format,"S2Q") != NULL);
    P2Q = (strstr(format,"P2Q") != NULL);
    P2R = (strstr(format,"P2R") != NULL);
    Q2P = (strstr(format,"Q2P") != NULL);

    if (ExA) key = key + 1;
    if (AxA) key = key + 2;
    if (ExB) key = key + 4;
    if (AxB) key = key + 8;
    if (ExR) key = key + 16;
    if (AxR) key = key + 32;
    if (ExS) key = key + 64;
    if (AxS) key = key + 128;
    if (ExP) key = key + 256;
    if (AxP) key = key + 512;
    if (ExQ) key = key + 1024;
    if (AxQ) key = key + 2048;

    if (A2B) key = key + 4096;
    if (R2S) key = key + 8192;
    if (R2P) key = key + 16384;
    if (R2Q) key = key + 32768;
    if (S2R) key = key + 65536;
    if (S2P) key = key + 131072;
    if (S2Q) key = key + 262144;
    if (P2Q) key = key + 524288;
    if (P2R) key = key + 1048576;
    if (Q2P) key = key + 2097152;

    h = Bdd_RelOp(f,g,format,FALSE);

  }

  if (!firsttime) {

    if ((f.p->count) && (f.p->count != bdd_counter)) {
      fprintf(stderr,
              "ERROR (Bdd_RelOp): Bad f (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(f.p)->count);
      exit(1);
    }

    if ((g.p->count) && (g.p->count != bdd_counter)) {
      fprintf(stderr,
             "ERROR (Bdd_RelOp): Bad g (bdd_counter = %d, count = %d)!\n",
              bdd_counter,(g.p)->count);
      exit(1);
    }

    if (op == oNULL) {
      if (Bdd_Cmp(f, bdd_termTrue) || Bdd_Cmp(f, bdd_termFalse))
         return f;
    }

    if (op == oAND) {
      if (Bdd_Cmp(f, bdd_termFalse) || Bdd_Cmp(g, bdd_termFalse))
        return bdd_termFalse;
      if (Bdd_Cmp(f, bdd_termTrue) && Bdd_Cmp(g, bdd_termTrue))
        return bdd_termTrue;
    }

    if (op == oOR) {
      if (Bdd_Cmp(f, bdd_termTrue) || Bdd_Cmp(g, bdd_termTrue))
        return bdd_termTrue;
      if (Bdd_Cmp(f, bdd_termFalse) && Bdd_Cmp(g, bdd_termFalse))
        return bdd_termFalse;
    }

    if (op == oIMPL) {
      if (Bdd_Cmp(f, bdd_termFalse) || Bdd_Cmp(g, bdd_termTrue))
        return bdd_termTrue;
      if (Bdd_Cmp(f, bdd_termTrue) && Bdd_Cmp(g, bdd_termFalse))
        return bdd_termFalse;
    }

    /* IF COMUTATIVE OPERATION, DO SIMPLE NORMALIZATION */
    if ((op == oOR) || (op == oAND)) {
      if ((unsigned int) f.p > (unsigned int) g.p) Bdd_Exchange(&f, &g);
    }

    /* LOOKUP IN CACHE TABLE */

    bddRelOpCache.search++;
    p = &bddRelOpCache.table[(
                    (unsigned int) key +
                    ((unsigned int) g.p << 15) +
		    (unsigned int) f.p +
		    ((unsigned int) g.mark << 1)+
		    (unsigned int) f.mark
		    ) % bddRelOpCache.size];

    if ((p->key == key) &&
        (p->f == f.p) &&
        (p->g == g.p) &&
	(p->fmark == f.mark) &&
        (p->gmark == g.mark))
    {
      bddRelOpCache.find++;
      h.p = p->r;
      h.mark = p->rmark;
      Bdd_Fresh(h.p);
      return h;
    }

    /* DETERMINING THE SMALLEST TOP VARIABLE AND THE SUCCESSORS */

    if (op == oNULL) {
      z = f.p->v;
      f0 = Bdd_TransferMark(f.p->f, f.mark);
      f1 = Bdd_TransferMark(f.p->t, f.mark);
    } else {
      if (f.p->v->order < g.p->v->order) {
        z = f.p->v;
        f0 = Bdd_TransferMark(f.p->f, f.mark);
        f1 = Bdd_TransferMark(f.p->t, f.mark);
        g0.p = g.p; g0.mark = g.mark; 
        g1.p = g.p; g1.mark = g.mark;
      } else if (f.p->v->order > g.p->v->order) {
        z = g.p->v;
        f0.p = f.p; f0.mark = f.mark;
        f1.p = f.p; f1.mark = f.mark;
        g0 = Bdd_TransferMark(g.p->f, g.mark);
        g1 = Bdd_TransferMark(g.p->t, g.mark);
      } else {
        z = f.p->v;
        f0 = Bdd_TransferMark(f.p->f, f.mark);
        f1 = Bdd_TransferMark(f.p->t, f.mark);
        g0 = Bdd_TransferMark(g.p->f, g.mark);
        g1 = Bdd_TransferMark(g.p->t, g.mark);
      }
    }

    /* DO RECURSIVE CALLS */

    if (op == oNULL) {
      h0 = Bdd_RelOp(f0, g, format, firsttime);
      h1 = Bdd_RelOp(f1, g, format, firsttime);
    } else {
      h0 = Bdd_RelOp(f0, g0, format, firsttime);
      h1 = Bdd_RelOp(f1, g1, format, firsttime);
    }

    /* COMPOSE PARTIAL RESULTS */
    switch (z->name[0]) {

    case 'a':
    {
      done = FALSE;

      if (ExA) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxA) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (A2B) {
        v = strdup(z->name);
        v[0] = 'b';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

      break;
    }

    case 'b':
    {
      done = FALSE;

      if (ExB) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxB) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

      break;
    }

    case 'r':
    {
      done = FALSE;

      if (ExR) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxR) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (R2S) {
        v = strdup(z->name);
        v[0] = 's';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (R2P) {
        v = strdup(z->name);
        v[0] = 'p';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (R2Q) {
        v = strdup(z->name);
        v[0] = 'q';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

      break;
    }

    case 's':
    {
      done = FALSE;

      if (ExS) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxS) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (S2R) {
        v = strdup(z->name);
        v[0] = 'r';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (S2P) {
        v = strdup(z->name);
        v[0] = 'p';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (S2Q) {
        v = strdup(z->name);
        v[0] = 'q';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

      break;
    }

    case 'p':
    {
      done = FALSE;

      if (ExP) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxP) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (P2Q) {
        v = strdup(z->name);
        v[0] = 'q';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (P2R) {
        v = strdup(z->name);
        v[0] = 'r';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

      break;
    }

    case 'q':
    {
      done = FALSE;

      if (ExQ) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxQ) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (Q2P) {
        v = strdup(z->name);
        v[0] = 'p';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

      break;
    }

    } /* switch z->name[0] */

    /* STORE IN CACHE TABLE */

    if (p->key) bddRelOpCache.overwrite++;
    p->f = f.p;   p->fmark = f.mark;
    p->g = g.p;   p->gmark = g.mark;
    p->r = h.p;   p->rmark = h.mark;
    p->key = key;

  }

  return h;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_RelOpComplex(Bdd_Edge f, Bdd_Edge g, Bdd_String format,
                 Bdd_Boolean firsttime)
{
  static int op;
  static unsigned int key;

  BddRelOpComplexCache *p;
  Bdd_Variable *z;
  Bdd_Edge f0, f1, g0, g1, h0, h1, h;

  static int n;            /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_String t;     /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_String zname; /* STATIC TO MINIMIZE STACK PROBLEMS */

  static Bdd_String v;     /* STATIC TO MINIMIZE STACK PROBLEMS */
  static Bdd_Boolean done; /* STATIC TO MINIMIZE STACK PROBLEMS */

  Bdd_Boolean Ex;
  Bdd_Boolean Ax;
  static Bdd_Boolean ExA;
  static Bdd_Boolean AxA;
  static Bdd_String ExR;
  static Bdd_String AxR;
  static Bdd_String ExS;
  static Bdd_String AxS;
  static Bdd_String ExP;
  static Bdd_String AxP;
  static Bdd_String ExQ;
  static Bdd_String AxQ;

  static Bdd_Boolean R2S;
  static Bdd_Boolean R2P;
  static Bdd_Boolean R2Q;
  static Bdd_Boolean S2R;
  static Bdd_Boolean S2P;
  static Bdd_Boolean S2Q;
  static Bdd_Boolean P2Q;
  static Bdd_Boolean Q2P;

  /* IF FIRSTIME, PARSING FORMAT STRING */

  if (firsttime) {

    if ((f.p->count) && (f.p->count != bdd_counter)) {
      fprintf(stderr,
        "ERROR (Bdd_RelOpComplex): Bad f (bdd_counter = %d, count = %d)!\n",
        bdd_counter,(f.p)->count);
      sleep(5);
      exit(1);
    }

    if ((g.p->count) && (g.p->count != bdd_counter)) {
      fprintf(stderr,
        "ERROR (Bdd_RelOpComplex): Bad g (bdd_counter = %d, count = %d)!\n",
        bdd_counter,(g.p)->count);
      sleep(5);
      exit(1);
    }

    key = 0;                              /* 0 * 2^30 */
    op = oNULL;

    if (strstr(format,"#AND")) {
      key = 1073741824U;                  /* 1 * 2^30 */
      op = oAND;
    }
  
    if (strstr(format,"#OR")) {
      key = 2147483648U;                  /* 2 * 2^30 */
      op = oOR;
    }

    if (strstr(format,"#IMPL")) {
      key = 3221225472U;                  /* 3 * 2^30 */
      op = oIMPL;
    }

    Ex = (strstr(format,"Ex") != NULL);
    Ax = (strstr(format,"Ax") != NULL);

    if (Ex) {

      ExA = (strstr(format,"xA") != NULL);

      if ((t = strstr(format,"xR"))) {
        n = strcspn(&t[2]," ");
        if (!(ExR = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(ExR,&t[2],n);
        ExR[n] = 0;

	/*
        ExR = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        ExR = NULL;
      }

      if ((t = strstr(format,"xS"))) {
        n = strcspn(&t[2]," ");
        if (!(ExS = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(ExS,&t[2],n);
        ExS[n] = 0;

	/*
        ExS = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        ExS = NULL;
      }

      if ((t = strstr(format,"xP"))) {
        n = strcspn(&t[2]," ");
        if (!(ExP = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(ExP,&t[2],n);
        ExP[n] = 0;
	
	/*
        ExP = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        ExP = NULL;
      }

      if ((t = strstr(format,"xQ"))) {
        n = strcspn(&t[2]," ");
        if (!(ExQ = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(ExQ,&t[2],n);
        ExQ[n] = 0;

	/*
        ExQ = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        ExQ = NULL;
      }

    } else {
      ExA = FALSE;
      ExR = NULL;
      ExS = NULL;
      ExP = NULL;
      ExQ = NULL;
    }

    if (Ax) {
      AxA = (strstr(format,"xA") != NULL);

      if ((t = strstr(format,"xR"))) {
        n = strcspn(&t[2]," ");
        if (!(AxR = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(AxR,&t[2],n);
        AxR[n] = 0;

	/*
        AxR = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        AxR = NULL;
      }

      if ((t = strstr(format,"xS"))) {
        n = strcspn(&t[2]," ");
        if (!(AxS = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(AxS,&t[2],n);
        AxS[n] = 0;

	/*
        AxS = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        AxS = NULL;
      }

      if ((t = strstr(format,"xP"))) {
        n = strcspn(&t[2]," ");
        if (!(AxP = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(AxP,&t[2],n);
        AxP[n] = 0;

	/*
        AxP = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        AxP = NULL;
      }

      if ((t = strstr(format,"xQ"))) {
        n = strcspn(&t[2]," ");
        if (!(AxQ = (Bdd_String) malloc(n+1))) {
          fprintf(stderr,"Bdd_RelOpComplex: Out of memory!\n");
          exit(1);
        }
        memcpy(AxQ,&t[2],n);
        AxQ[n] = 0;

	/*
        AxQ = (Bdd_String) strtok(strdup(&t[2]), " ");
	*/
      } else {
        AxQ = NULL;
      }

    } else {
      AxA = FALSE;
      AxR = NULL;
      AxS = NULL;
      AxP = NULL;
      AxQ = NULL;
    }

    R2S = (strstr(format,"R2S") != NULL);
    R2P = (strstr(format,"R2P") != NULL);
    R2Q = (strstr(format,"R2Q") != NULL);
    S2R = (strstr(format,"S2R") != NULL);
    S2P = (strstr(format,"S2P") != NULL);
    S2Q = (strstr(format,"S2Q") != NULL);
    P2Q = (strstr(format,"P2Q") != NULL);
    Q2P = (strstr(format,"Q2P") != NULL);

    if (ExA) key = key + 1;
    if (AxA) key = key + 2;
    if (ExR) key = key + 4;
    if (AxR) key = key + 8;
    if (ExS) key = key + 16;
    if (AxS) key = key + 32;
    if (ExP) key = key + 64;
    if (AxP) key = key + 128;
    if (ExQ) key = key + 256;
    if (AxQ) key = key + 512;
 
    if (R2S) key = key + 1024;
    if (R2P) key = key + 2048;
    if (R2Q) key = key + 4096;
    if (S2R) key = key + 8192;
    if (S2P) key = key + 16384;
    if (S2Q) key = key + 32768;
    if (P2Q) key = key + 65536;
    if (Q2P) key = key + 131072;

    if (ExR && strcmp(ExR,"*")) {
      key = key + 4194304;
    }

    if (AxR && strcmp(AxR,"*")) {
      key = key + 4194304;
    }

    if (ExS && strcmp(ExS,"*")) {
      key = key + 8388608;
    }

    if (AxS && strcmp(AxS,"*")) {
      key = key + 8388608;
    }

    if (ExP && strcmp(ExP,"*")) {
      key = key + 16777216;
    }

    if (AxP && strcmp(AxP,"*")) {
      key = key + 16777216;
    }

    if (ExQ && strcmp(ExQ,"*")) {
      key = key + 33554432;
    }

    if (AxQ && strcmp(AxQ,"*")) {
      key = key + 33554432;
    }

    h = Bdd_RelOpComplex(f,g,format,FALSE);

    if (ExR) free(ExR);
    if (AxR) free(AxR);
    if (ExS) free(ExS);
    if (AxS) free(AxS);
    if (ExP) free(ExP);
    if (AxP) free(AxP);
    if (ExQ) free(ExQ);
    if (AxQ) free(AxQ);
  }

  if (!firsttime) {

    if (op == oNULL) {
      if (Bdd_Cmp(f, bdd_termTrue) || Bdd_Cmp(f, bdd_termFalse))
        return f;
    }

    if (op == oAND) {
      if (Bdd_Cmp(f, bdd_termFalse) || Bdd_Cmp(g, bdd_termFalse))
        return bdd_termFalse;
      if (Bdd_Cmp(f, bdd_termTrue) && Bdd_Cmp(g, bdd_termTrue))
        return bdd_termTrue;
    }

    if (op == oOR) {
      if (Bdd_Cmp(f, bdd_termTrue) || Bdd_Cmp(g, bdd_termTrue))
        return bdd_termTrue;
      if (Bdd_Cmp(f, bdd_termFalse) && Bdd_Cmp(g, bdd_termFalse))
        return bdd_termFalse;
    }

    if (op == oIMPL) {
      if (Bdd_Cmp(f, bdd_termFalse) || Bdd_Cmp(g, bdd_termTrue))
        return bdd_termTrue;
      if (Bdd_Cmp(f, bdd_termTrue) && Bdd_Cmp(g, bdd_termFalse))
        return bdd_termFalse;
    }

    /* IF COMUTATIVE OPERATION, DO SIMPLE NORMALIZATION */
    if ((op == oOR) || (op == oAND)) {
      if ((unsigned int) f.p > (unsigned int) g.p) Bdd_Exchange(&f, &g);
    }

    /* LOOKUP IN CACHE TABLE */

    bddRelOpComplexCache.search++;
    p = &bddRelOpComplexCache.table[(
                    (unsigned int) key +
                    ((unsigned int) g.p << 15) +
		    (unsigned int) f.p +
		    ((unsigned int) g.mark << 1) +
		    (unsigned int) f.mark
		    ) % bddRelOpComplexCache.size];

    if (p->key == key &&
        p->f == f.p && p->g == g.p &&
	p->fmark == f.mark && p->gmark == g.mark &&
        !strcmp(p->name,format))
    {
      bddRelOpComplexCache.find++;
      h.p = p->r;
      h.mark = p->rmark;
      Bdd_Fresh(h.p);
      return h;
    }

    /* DETERMINING THE SMALLEST TOP VARIABLE AND THE SUCCESSORS */

    if (op == oNULL) {
      z = f.p->v;
      f0 = Bdd_TransferMark(f.p->f, f.mark);
      f1 = Bdd_TransferMark(f.p->t, f.mark);
    } else {
      if (f.p->v->order < g.p->v->order) {
        z = f.p->v;
        f0 = Bdd_TransferMark(f.p->f, f.mark);
        f1 = Bdd_TransferMark(f.p->t, f.mark);
        g0.p = g.p; g0.mark = g.mark; 
        g1.p = g.p; g1.mark = g.mark;
      } else if (f.p->v->order > g.p->v->order) {
        z = g.p->v;
        f0.p = f.p; f0.mark = f.mark;
        f1.p = f.p; f1.mark = f.mark;
        g0 = Bdd_TransferMark(g.p->f, g.mark);
        g1 = Bdd_TransferMark(g.p->t, g.mark);
      } else {
        z = f.p->v;
        f0 = Bdd_TransferMark(f.p->f, f.mark);
        f1 = Bdd_TransferMark(f.p->t, f.mark);
        g0 = Bdd_TransferMark(g.p->f, g.mark);
        g1 = Bdd_TransferMark(g.p->t, g.mark);
      }
    }

    /* DO RECURSIVE CALLS */

    if (op == oNULL) {
      h0 = Bdd_RelOpComplex(f0, g, format, firsttime);
      h1 = Bdd_RelOpComplex(f1, g, format, firsttime);
    } else {
      h0 = Bdd_RelOpComplex(f0, g0, format, firsttime);
      h1 = Bdd_RelOpComplex(f1, g1, format, firsttime);
    }

    /* FIND OUT THE NAME OF THE PROCESS */

    if (z->name[0] != 'a') {
      zname = &z->name[1+strspn(&z->name[1],"0123456789")];
    }

    /* COMPOSE PARTIAL RESULTS */

    if (z->name[0] == 'a') {
      done = FALSE;

      if (ExA) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxA) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

    } /* z->name[0] == 'a' */

    if (z->name[0] == 'r') {
      done = FALSE;

      if (ExR && (!strcmp(ExR,"*") || strstr(ExR,zname))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxR && (!strcmp(AxR,"*") || strstr(AxR,zname))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (R2S) {
        v = strdup(z->name);
        v[0] = 's';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (R2P) {
        v = strdup(z->name);
        v[0] = 'p';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (R2Q) {
        v = strdup(z->name);
        v[0] = 'q';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

    } /* z->name[0] == 'r' */
  
    if (z->name[0] == 's') {
      done = FALSE;

      if (ExS && (!strcmp(ExS,"*") || strstr(ExS,zname))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxS && (!strcmp(AxS,"*") || strstr(AxS,zname))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (S2R) {
        v = strdup(z->name);
        v[0] = 'r';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (S2P) {
        v = strdup(z->name);
        v[0] = 'p';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (S2Q) {
        v = strdup(z->name);
        v[0] = 'q';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

    } /* z->name[0] == 's' */

    if (z->name[0] == 'p') {
      done = FALSE;

      if (ExP && (!strcmp(ExP,"*") || strstr(ExP,zname))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxP && (!strcmp(AxP,"*") || strstr(AxP,zname))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (P2Q) {
        v = strdup(z->name);
        v[0] = 'q';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

    } /* z->name[0] == 'p' */

    if (z->name[0] == 'q') {
      done = FALSE;

      if (ExQ && (!strcmp(ExQ,"*") || strstr(ExQ,zname))) {
        h = Bdd_ITE(h0, bdd_termTrue, h1);
        done = TRUE;
      }

      if (AxQ && (!strcmp(AxQ,"*") || strstr(AxQ,zname))) {
        h = Bdd_ITE(h0, h1, bdd_termFalse);
        done = TRUE;
      }

      if (Q2P) {
        v = strdup(z->name);
        v[0] = 'p';
        h = Bdd_ITE(Bdd_FoaTerminal(v), h1, h0);
        free(v);
        done = TRUE;
      }

      if (!done) {
        h =  Bdd_FoaNode(z, h0, h1);
      }

    } /* z->name[0] == 'q' */


    /* STORE IN CACHE TABLE */

    if (p->key) {
      bddRelOpComplexCache.overwrite++;
      free(p->name);
    }
    p->f = f.p;   p->fmark = f.mark;
    p->g = g.p;   p->gmark = g.mark;
    p->r = h.p;   p->rmark = h.mark;
    p->key = key;
    p->name = strdup(format);

  }

  return h;
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs existential quantification over 'a' variables
               in product f*g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExA(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#AND Ex xA",TRUE);
}


/**Function****************************************************************
  Synopsis    []
  Description [Performs existential quantification over 'r' variables
               in product f*g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExR(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#AND Ex xR",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs existential quantification over 's' variables
               in product f*g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExS(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#AND Ex xS",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs existential quantification over 'q' variables
               in product f*g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExQ(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#AND Ex xQ",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs existential quantification over 'p' variables
               in product f*g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExP(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#AND Ex xP",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs universal quantification over 'a' variables
               in sum f+g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_AxA(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#OR Ax xA",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs universal quantification over 'r' variables
               in sum f+g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_AxR(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#OR Ax xR",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs universal quantification over 's' variables
               in sum f+g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_AxS(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#OR Ax xS",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs universal quantification over 'p' variables
               in sum f+g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_AxP(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#OR Ax xP",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs existential quantification over all variables,
               except 'r', in product f*g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_ExR_1(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#AND Ex xA xS xP xQ",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description [Performs universal quantification over all variables,
               except 'r', in sum f+g.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_AxR_1(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#OR Ax xA xS xP xQ",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['r' variables replaces with 's'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_R2S(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"R2S",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['s' variables replaces with 'r'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_S2R(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"S2R",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['p' variables replaces with 'q'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_P2Q(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"P2Q",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['s' variables replaces with 'p'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_S2P(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"S2P",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['r' variables replaces with 'p'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_R2P(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"R2P",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['r' variables replaces with 'q'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_R2Q(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"R2Q",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['r' variables replaces with 'p'.
               's' variables replaces with 'q'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_R2P_S2Q(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"R2P S2Q",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['r' variables replaces with 's'.
               'p' variables replaces with 'q'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_R2S_P2Q(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"R2S P2Q",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['r' variables replaces with 'p'.
               'p' variables replaces with 'r'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_R2P_P2R(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"R2P P2R",TRUE);
}

/**Function****************************************************************
  Synopsis    []
  Description ['s' variables replaces with 'r'.
               'q' variables replaces with 'p'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_S2R_Q2P(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"S2R Q2P",TRUE);
}

/**Function****************************************************************
  Synopsis    [NewState]
  Description ['s' variables replaces with 'r'. Performs existential
               quantification over all other variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewState(Bdd_Edge f)
{
  return Bdd_RelOp(f,bdd_termFalse,"S2R Ex xR xA xP xQ",TRUE);
}

/**Function****************************************************************
  Synopsis    [NewState1]
  Description ['s' variables replaces with 'r'.
               Performs existential quatification over 'r' variables
               of the given process and existential quatification over
               other variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewState1(Bdd_Edge f, Bdd_String name)
{
  Bdd_String s;
  Bdd_Edge r;

  s = (Bdd_String) malloc(255);
  sprintf(s,"S2R Ex xA xR<%s> xP* xQ*",name);
  r = Bdd_RelOpComplex(f,bdd_termFalse,s,TRUE);
  free(s);
  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description ['s' variables replaces with 'r'.
               Performs existential quantification over 'r' variables
               of the given two processes and existential quantification
               over other variables.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewState2(Bdd_Edge f, Bdd_String name1, Bdd_String name2)
{
  Bdd_String s;
  Bdd_Edge r;

  s = (Bdd_String) malloc(255);
  sprintf(s,"S2R Ex xA xR<%s><%s> xP* xQ*",name1,name2);
  r = Bdd_RelOpComplex(f,bdd_termFalse,s,TRUE);
  free(s);
  return r;
}

/**Function****************************************************************
  Synopsis    []
  Description [Calculate f->g. Performs universal quantification over
               'a', 's' and 'q'.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

Bdd_Edge
Bdd_NewEq(Bdd_Edge f, Bdd_Edge g)
{
  return Bdd_RelOp(f,g,"#IMPL Ax xA xS xQ",TRUE);
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function BddRelOpGarbage.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddRelOpGarbage()
{
  int i;
  BddRelOpCache *p;

  for (i=0; i<bddRelOpCache.size; i++) {
    if (bddRelOpCache.table[i].key) {
      p = &bddRelOpCache.table[i];
      if ((p->f->count && p->f->count != bdd_counter) ||
          (p->g->count && p->g->count != bdd_counter) ||
          (p->r->count && p->r->count != bdd_counter))
      {
        p->key = 0;
      }
    }
  }
}

/**Function****************************************************************
  Synopsis    [Function BddRelOpComplexGarbage.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

void
BddRelOpComplexGarbage()
{
  int i;
  BddRelOpComplexCache *p;

  for (i=0; i<bddRelOpComplexCache.size; i++) {
    if (bddRelOpComplexCache.table[i].key) {
      p = &bddRelOpComplexCache.table[i];
      if ((p->f->count && p->f->count != bdd_counter) ||
          (p->g->count && p->g->count != bdd_counter) ||
          (p->r->count && p->r->count != bdd_counter))
      {
        free(p->name);
        p->key = 0;
      }
    }
  }
}

