/**CFile*******************************************************************
  PackageName [bdd]
  FileName    [bddTcl.c]
  Authors     [Robert Meolic (meolic@uni-mb.si),
               Ales Casar (casar@uni-mb.si)]

  Revision    [$Revision: 2.8 $]
  Date        [$Date: 2004/07/22 14:32:04 $]
  Synopsis    [Package 'bdd' provides data structures and algorithms
               for the representation and manipulation of Boolean
               functions with ROBDDs. A hash table is used for quick
               search of nodes. Complement edges decreases the number
               of nodes. Function ITE is given for manipulation of
               boolean functions. An automatic garbage collection with
               a formulae counter is implemented.]
  Description [The file bddTcl.c contains definitions of Tcl commands,
               which can be used for manipulating with BDDs from
               Tcl interpreter (e.g. tclsh or wish).]
  SeeAlso     [bdd.h, bddInt.h]
  Copyright   [Copyright (C) 2000-2004 Robert Meolic
               This program is free software; you can redistribute it
               and/or modify it under the terms of the GNU General Public
               License (version 2) as published by the Free Software
               Foundation.

               This program is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 59 Temple Place, Suite 330,
               Boston, MA  02111-1307 USA.]
  ************************************************************************/

#include "bddInt.h"

/* on tcl 8.3 use #define USECONST */
/* on tcl 8.4 use #define USECONST const*/
#define USECONST const

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Static function prototypes                                            */
/*-----------------------------------------------------------------------*/

static int BddInitPkgCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddExitPkgCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddAboutPkgCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTerminalsCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddExprCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddNOTCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddITECmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddECmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddACmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddRestrictCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddComposeCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddOutFunctionCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddOutBDDCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddCmpCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddFindCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddNodeNumberCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddNodeMaxLevelCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddNodeAvgLevelCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableSizeCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableNumberCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableMaxCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableFoaCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableAddCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableCompareCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableGeneratedCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableFortifiedCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableFreshCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddTableBadCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddWriteTableCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddListUsedCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddListMaxLengthCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddListAvgLengthCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddCacheIteSizeCmd(ClientData clientData, Tcl_Interp *interp,
                              int argc, USECONST char **argv);

static int BddCacheIteOkCmd(ClientData clientData, Tcl_Interp *interp,
                            int argc, USECONST char **argv);

static int BddGarbageNumberCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddFindNodeCmd(ClientData clientData, Tcl_Interp *interp,
                          int argc, USECONST char **argv);

static int BddCheckTableCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, USECONST char **argv);

static int BddWriteTreeCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, USECONST char **argv);

static int BddTreeMaxLevelCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, USECONST char **argv);

static int BddTreeMinLevelCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, USECONST char **argv);

static int BddTreeAvgLevelCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, USECONST char **argv);

static int BddTreeNumberCmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, USECONST char **argv);

static int wordNumber(Bdd_String s);

/**AutomaticEnd***********************************************************/

/*-----------------------------------------------------------------------*/
/* Definition of exported functions                                      */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function BDD_Init.]
  Description [This function is used by Tcl/Tk only.]
  SideEffects []
  SeeAlso     []
  ************************************************************************/

int
Bdd_Init(Tcl_Interp *interp)
{
  Tcl_CreateCommand(interp, "bdd_initPkg", BddInitPkgCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_exitPkg", BddExitPkgCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_aboutPkg", BddAboutPkgCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_terminals", BddTerminalsCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_expr", BddExprCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_NOT", BddNOTCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_ITE", BddITECmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_E", BddECmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_A", BddACmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_restrict", BddRestrictCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_compose", BddComposeCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_out_function", BddOutFunctionCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_out_BDD", BddOutBDDCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_cmp", BddCmpCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_find", BddFindCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_node_number", BddNodeNumberCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_node_max_level", BddNodeMaxLevelCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_node_avg_level", BddNodeAvgLevelCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_size", BddTableSizeCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_number", BddTableNumberCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_max", BddTableMaxCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_foa", BddTableFoaCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_add", BddTableAddCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_compare", BddTableCompareCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_generated", BddTableGeneratedCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_fortified", BddTableFortifiedCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_fresh", BddTableFreshCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_table_bad", BddTableBadCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_write_table", BddWriteTableCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_list_used", BddListUsedCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_list_max_length", BddListMaxLengthCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_list_avg_length", BddListAvgLengthCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_cache_ITE_size", BddCacheIteSizeCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_cache_ITE_ok", BddCacheIteOkCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_garbage_number", BddGarbageNumberCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* ************************************************ */
  /* THE FOLLOWING COMMANDS ARE USED ONLY FOR TESTING */
  /* ************************************************ */

  Tcl_CreateCommand(interp, "bdd_find_node", BddFindNodeCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_check_table", BddCheckTableCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_write_tree", BddWriteTreeCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_tree_max_level", BddTreeMaxLevelCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_tree_min_level", BddTreeMinLevelCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_tree_avg_level", BddTreeAvgLevelCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand(interp, "bdd_tree_number", BddTreeNumberCmd,
                     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  return Tcl_PkgProvide(interp, "bdd", "1.0");
}

/*-----------------------------------------------------------------------*/
/* Definition of internal functions                                      */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Definition of static functions                                        */
/*-----------------------------------------------------------------------*/

/**Function****************************************************************
  Synopsis    [Function BddInitCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddInitPkgCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                   USECONST char **argv)
{
  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  printf("Initialization of BDD package... ");
  Bdd_InitPkg();
  printf("OK");

  printf("\n");
  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddExitCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddExitPkgCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                   USECONST char **argv)
{
  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  printf("Exit BDD package... ");
  Bdd_ExitPkg();
  printf("OK");

  printf("\n");
  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddAboutPkgCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddAboutPkgCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  Bdd_AboutPkg();

  printf("\n");
  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTerminalsCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTerminalsCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                     USECONST char **argv)
{
  Bdd_String s1;
  Bdd_Edge sup;
  Bdd_String name;
  int n,k;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  n = wordNumber(s1);

  name = strtok(s1," ");  /* first terminal name */
  sup = Bdd_FoaTerminal(name);
  Bdd_AddFormula(name,sup);

  for(k=1;k<n; k++){
    name = strtok(NULL," ");  /* next terminals name */
    sup = Bdd_FoaTerminal(name);
    Bdd_AddFormula(name,sup);
  }

  free(s1);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddExprCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddExprCmd(ClientData clientData, Tcl_Interp *interp, int argc,
               USECONST char **argv)
{
  Bdd_String s1,s2;

  if (argc != 3) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);

  Bdd_Expr(s1,s2);

  free(s1);
  free(s2);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddNOTCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddNOTCmd(ClientData clientData, Tcl_Interp *interp, int argc,
               USECONST char **argv)
{
  Bdd_String s1,s2;
  Bdd_Edge f,r;

  if (argc != 3){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);

  if (Bdd_FindFormula(s1,&f)) {
        
    r = Bdd_NOT(f);
    Bdd_AddFormula(s2,r);

  } else printf("Missing Boolean function %s.\n",s1);

  free(s1);
  free(s2);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddITECmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddITECmd(ClientData clientData, Tcl_Interp *interp, int argc,
               USECONST char **argv)
{
  Bdd_String s1,s2,s3,s4;
  Bdd_Edge f,g,h,r;

  if (argc != 5){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);
  s3 = strdup(argv[3]);
  s4 = strdup(argv[4]);

  if ( Bdd_FindFormula(s1,&f) && Bdd_FindFormula(s2,&g) &&
       Bdd_FindFormula(s3,&h) ) {
        
    r = Bdd_ITE(f,g,h);
    Bdd_AddFormula(s4,r);

  } else printf("Missing Boolean function.\n");

  free(s1);
  free(s2);
  free(s3);
  free(s4);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddECmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddECmd(ClientData clientData, Tcl_Interp *interp, int argc,
             USECONST char **argv)
{
  Bdd_String s1,s2,s3;
  Bdd_Edge f,sup;
  Bdd_String name;
  int n,k;

  if (argc != 4){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);
  s3 = strdup(argv[3]);

  if (Bdd_FindFormula(s2,&f)) {

    n = wordNumber(s1);

    name = strtok(s1," ");  /* first terminal name */
    sup = Bdd_FindTerminal(name);
    if (sup.p != bdd_termNull.p) {
      f = Bdd_E(f,sup.p->v);
    } else printf("Non-existing variable.\n");

    for(k=1;k<n; k++){
      name = strtok(NULL," ");  /* next terminals name */
      sup = Bdd_FindTerminal(name);
      if (sup.p != bdd_termNull.p) {
        f = Bdd_E(f,sup.p->v);
      } else printf("Non-existing variable.\n");
    }

    Bdd_AddFormula(s3,f);

  } else printf("Non-existing Boolean function.\n");

  free(s1);
  free(s2);
  free(s3);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddACmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddACmd(ClientData clientData, Tcl_Interp *interp, int argc,
             USECONST char **argv)
{
  Bdd_String s1,s2,s3;
  Bdd_Edge f,sup;
  Bdd_String name;
  int n,k;

  if (argc != 4){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);
  s3 = strdup(argv[3]);

  if (Bdd_FindFormula(s2,&f)) {

    n = wordNumber(s1);

    name = strtok(s1," ");  /* first terminal name */
    sup = Bdd_FindTerminal(name);
    if (sup.p != bdd_termNull.p) {
      f = Bdd_A(f,sup.p->v);
    } else printf("Non-existing variable.\n");
  
    for(k=1;k<n; k++){
      name = strtok(NULL," ");  /* next terminals name */
      sup = Bdd_FindTerminal(name);
      if (sup.p != bdd_termNull.p) {
        f = Bdd_A(f,sup.p->v);
      } else printf("Non-existing variable.\n");
    }

    Bdd_AddFormula(s3,f);

  } else printf("Non-existing Boolean function.\n");

  free(s1);
  free(s2);
  free(s3);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddRestrictCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddRestrictCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  Bdd_String s1,s2,s3,s4;
  Bdd_Edge f,sup;
  int b;

  if (argc != 5){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);
  s3 = strdup(argv[3]);
  s4 = strdup(argv[4]);

  sscanf(s3,"%d",&b);

  if (Bdd_FindFormula(s1,&f)) {

    sup = Bdd_FindTerminal(s2);
    if (sup.p != bdd_termNull.p) {
      if (b == 0) f = Bdd_Restrict(f,sup.p->v,FALSE);
        else f = Bdd_Restrict(f,sup.p->v,TRUE);
    } else printf("Missing variable.\n");
  
    Bdd_AddFormula(s4,f);

  } else printf("Missing Boolean function.\n");

  free(s1);
  free(s2);
  free(s3);
  free(s4);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddComposeCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddComposeCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                   USECONST char **argv)
{
  Bdd_String s1,s2,s3,s4;
  Bdd_Edge f,g,r,sup;

  if (argc != 5){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);
  s3 = strdup(argv[3]);
  s4 = strdup(argv[4]);

  if ( Bdd_FindFormula(s1,&f) && Bdd_FindFormula(s3,&g) ) {

    sup = Bdd_FindTerminal(s2);
    if (sup.p != bdd_termNull.p) {
      r = Bdd_Compose(f,sup.p->v,g);
    } else {
      r = bdd_termNull;
      printf("Missing variable.\n");
    }

    Bdd_AddFormula(s4,r);

  } else printf("Missing Boolean function.\n");

  free(s1);
  free(s2);
  free(s3);
  free(s4);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddOutFunctionCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddOutFunctionCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                       USECONST char **argv)
{
  Bdd_String s1;
  Bdd_Edge f;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  if (Bdd_FindFormula(s1,&f)) {

    printf("Boolean function %s:\n",s1); 
    Bdd_WriteFunction(f);

  } else printf("Missing Boolean function.\n");

  free(s1);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddOutBDDCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddOutBDDCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                  USECONST char **argv)
{
  Bdd_String s1;
  Bdd_Edge f;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  if (Bdd_FindFormula(s1,&f)) {

    printf("BDD for Boolean function %s:\n",s1); 
    Bdd_WriteBDD(f);

  } else printf("Missing Boolean function.\n");

  free(s1);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}


/**Function****************************************************************
  Synopsis    [Function BddCmpCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddCmpCmd(ClientData clientData, Tcl_Interp *interp, int argc,
               USECONST char **argv)
{
  Bdd_String s1,s2;
  Bdd_Edge f,g;
  Bdd_String sup;

  if (argc != 3){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);

  if ( Bdd_FindFormula(s1,&f) && Bdd_FindFormula(s2,&g) ) {
        
    if (Bdd_Cmp(f,g)) {
      sup = (Bdd_String) strdup("TRUE");
    } else {
      sup = (Bdd_String) strdup("FALSE");
    }

  } else printf("Missing Boolean function.\n");

  free(s1);
  free(s2);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddFindCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddFindCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                USECONST char **argv)
{
  Bdd_String s1;
  Bdd_Edge f;
  Bdd_String sup;

  if (argc != 2){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  if ( Bdd_FindFormula(s1,&f)) {
    sup = (Bdd_String) strdup("TRUE");
  } else {
    sup = (Bdd_String) strdup("FALSE");
  }

  free(s1);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddNodeNumberCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddNodeNumberCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                      USECONST char **argv)
{
  Bdd_String s1;
  int i;
  Bdd_Edge f;
  Bdd_String sup;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  if (Bdd_FindFormula(s1,&f)) {

    i = Bdd_NodeNumber(f.p);
    sup = (Bdd_String) malloc(15);
    sprintf(sup,"%d",i);

  } else {
    printf("Missing Boolean function.\n");
    sup = "";
  }

  free(s1);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddNodeMaxLevelCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddNodeMaxLevelCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  Bdd_String s1;
  int i;
  Bdd_Edge f;
  Bdd_String sup;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  if (Bdd_FindFormula(s1,&f)) {

    i = Bdd_NodeMaxLevel(f.p);
    sup = (Bdd_String) malloc(15);
    sprintf(sup,"%d",i);

  } else {
    printf("Missing Boolean function.\n");
    sup = "";
  }

  free(s1);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddNodeAvgLevelCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddNodeAvgLevelCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  Bdd_String s1;
  float d;
  Bdd_Edge f;
  Bdd_String sup;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  if (Bdd_FindFormula(s1,&f)) {

    d = Bdd_NodeAvgLevel(f.p);
    sup = (Bdd_String) malloc(15);
    sprintf(sup,"%.2f",d);

  } else {
    printf("Missing Boolean function.\n");
    sup = "";
  }

  free(s1);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableSizeCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableSizeCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                     USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.size);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}


/**Function****************************************************************
  Synopsis    [Function BddTableNumberCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableNumberCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                       USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.num);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableMaxCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableMaxCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.max);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableFoaCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableFoaCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.foa);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableAddCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableAddCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.add);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableCompareCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableCompareCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.compare);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableGeneratedCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableGeneratedCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                          USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.generated);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableFortifiedCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableFortifiedCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                          USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.nf);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableFreshCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableFreshCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                      USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%s","NOT IMPLEMENTED");
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTableBadCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTableBadCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%s","NOT IMPLEMENTED");
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddWriteTableCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddWriteTableCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                      USECONST char **argv)
{
  Bdd_String s1;
  int i;

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);

  sscanf(s1,"%d",&i);
  BddWriteTable(i);

  free(s1);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddListUsedCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddListUsedCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  int i;
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  i = Bdd_ListUsed();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",i);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddListMaxLengthCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddListMaxLengthCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                         USECONST char **argv)
{
  int i;
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  i = Bdd_ListMaxLength();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",i);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddListAvgLengthCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddListAvgLengthCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                         USECONST char **argv)
{
  float d;
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  d = Bdd_ListAvgLength();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%.2f",d);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddHashIteSizeCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddCacheIteSizeCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddIteCache.size);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddListAvgLengthCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddCacheIteOkCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                      USECONST char **argv)
{
  int n;
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  n = Bdd_CacheIteOk();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",n);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddGarbageNumberCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddGarbageNumberCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                         USECONST char **argv)
{
  Bdd_String sup;

  if (argc != 1) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",bddNodeTable.garbage);
  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddFindNodeCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddFindNodeCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                    USECONST char **argv)
{
  Bdd_String s1,s2,s3;
  int f,t;
  Bdd_Variable *v;
  Bdd_Edge pv;

  if (argc != 4){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  s1 = strdup(argv[1]);
  s2 = strdup(argv[2]);
  s3 = strdup(argv[3]);

  pv = Bdd_FindTerminal(s1);
  v = pv.p->v;
  
  sscanf(s2,"%x",&f);
  sscanf(s3,"%x",&t);

  BddFindNode(v, (Bdd_Node *) f, (Bdd_Node *) t);

  free(s1);
  free(s2);
  free(s3);

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddCheckTableCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddCheckTableCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                      USECONST char **argv)
{
  if (argc != 1){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  BddCheckTable();

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddWriteTreeCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddWriteTreeCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                     USECONST char **argv)
{
  if (argc != 1){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  BddWriteTree();

  Tcl_SetResult(interp, "", TCL_STATIC);
  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTreeMaxLevelCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTreeMaxLevelCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  int i;
  Bdd_String sup;

  if (argc != 1){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  i = BddTreeMaxLevel();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",i);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTreeMinLevelCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTreeMinLevelCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  int i;
  Bdd_String sup;

  if (argc != 1){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  i = BddTreeMinLevel();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",i);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTreeAvgLevelCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTreeAvgLevelCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                        USECONST char **argv)
{
  float d;
  Bdd_String sup;

  if (argc != 1){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  d = BddTreeAvgLevel();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%.2f",d);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    [Function BddTreeNumberCmd.]
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
BddTreeNumberCmd(ClientData clientData, Tcl_Interp *interp, int argc,
                      USECONST char **argv)
{
  int i;
  Bdd_String sup;

  if (argc != 1){
    interp->result = "wrong # args";
    return TCL_ERROR;
  }

  i = BddTreeNumber();
  sup = (Bdd_String) malloc(15);
  sprintf(sup,"%d",i);

  Tcl_SetResult(interp, sup, TCL_VOLATILE);
  free(sup);

  return TCL_OK;
}

/**Function****************************************************************
  Synopsis    []
  Description []
  SideEffects []
  SeeAlso     []
  ************************************************************************/

static int
wordNumber(Bdd_String s)
{
  int num;
  int token;
  Bdd_String sup,sup_next;

  num = 0;
  sup = strdup(s);

  token = strspn(sup," ");
  while (strspn(sup," ") < strlen(sup)) {
    sup_next = (Bdd_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    num++;
    token = strcspn(sup," ");
    sup_next = (Bdd_String) strdup(&sup[token]);
    free(sup);
    sup = sup_next;

    token = strspn(sup," ");
  }

  free(sup);
  return num;
}
