/**CFile*******************************************************************
  PackageName [pa]
  FileName    [pa.h]
  Authors     [Robert Meolic (meolic@uni-mb.si)]

  Revision    [$Revision: 2.15 $]
  Date        [$Date: 2005/08/03 07:10:48 $]
  Synopsis    [Package 'pa' implements process algebra similar to Milner's
               CCS and Hoare's CSP.]
  Description [The file pa.h contains declaration of all external
               data structures neded for implementation of
               process algebra.]
  SeeAlso     [paInt.h, paMain.c]
  Copyright   [Copyright (C) 2000-2005 Robert Meolic
               This program is free software; you can redistribute it
               and/or modify it under the terms of the GNU General Public
               License (version 2) as published by the Free Software
               Foundation.

               This program is distributed in the hope that it will be useful,
               but WITHOUT ANY WARRANTY; without even the implied warranty of
               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
               GNU General Public License for more details.

               You should have received a copy of the GNU General Public
               License along with this program; if not, write to the Free
               Software Foundation, Inc., 59 Temple Place, Suite 330,
               Boston, MA  02111-1307 USA.]
  ************************************************************************/

#ifndef _PA
#define _PA

#define PA_RELEASE "2.3"

#include <pthread.h>
#include <sched.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include <unistd.h>
#include <math.h>

#include <bdd.h>

/* ON WIN32 THERE HAS TO BE DEFINED ESTWIN32 */
/* ON UNIX THERE HAS TO BE DEFINED ESTUNIX */

#ifdef ESTUNIX
#  ifndef EXTERN
#    define EXTERN extern
#  endif
#endif

#ifdef ESTWIN32
#ifndef TCL
#    ifdef BUILD_PA
#      undef EXTERN
#      define EXTERN __declspec (dllexport)
#    else
#      ifdef USE_PA
#        undef EXTERN
#        define EXTERN __declspec (dllimport)
#      else
#        define EXTERN extern
#      endif
#    endif
#else
#    ifdef BUILD_PA
#      undef TCL_STORAGE_CLASS
#      define TCL_STORAGE_CLASS DLLEXPORT
#    else
#      undef TCL_STORAGE_CLASS
#      define TCL_STORAGE_CLASS DLLIMPORT
#    endif
#endif
#endif

/*-----------------------------------------------------------------------*/
/* Constant declarations                                                 */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* Type declarations                                                     */
/*-----------------------------------------------------------------------*/

#ifndef _COMMONTYPE
#define _COMMONTYPE
typedef void *pvoid;
#endif

/*-----------------------------------------------------------------------*/
/* Structure declarations                                                */
/*-----------------------------------------------------------------------*/

typedef struct {
  Bdd_String  name;            /* action name */
  Bdd_Edge    p;               /* BDD representation of an action */
} Pa_Action;

typedef struct {
  Bdd_String  name;            /* state name */
  Bdd_Edge    bddR ;           /* BDD with 'r' variables (current state) */
  Bdd_Edge    bddS;            /* BDD with 's' variables (next state) */
  Bdd_Edge    bddComp;         /* represents how the state is composed */
  Bdd_Edge    acc;             /* acceptance set */
  Bdd_Edge    minacc;          /* minimized acceptance set */
  Bdd_Boolean final;           /* TRUE for final states, used for w/c automata */
  Bdd_String  ccs;             /* outgoing transition in CCS format */
} Pa_State;

typedef struct {
  int source;                             /* source state */
  int action;                             /* action */
  int type;                               /* action type, 0 = output, 1 = input */
  int dest;                               /* destination state */
} Pa_Transition;

typedef struct {
  Bdd_String   name;                         /* sort name */
  Bdd_Boolean  deleted;                      /* TRUE = deleted */
  int          encoded;                      /* 1 = yes, 0 = no */
  int          numVar;                       /* number of variables */
  int          numActions;                   /* number of actions */
  Pa_Action    *table;                       /* table of actions */
  Bdd_Edge     a0;                           /* first variable */
} Pa_Sort;

typedef struct {
  Bdd_String     name;                       /* process name */
  Bdd_Boolean    deleted;                    /* TRUE = deleted */
  Bdd_String     compname;                   /* optionally, if states are composed */
  int            encoded;                    /* 1 = yes, 0 = no */
  int            sort;                       /* sort */
  int            numStates;                  /* number of states */
  Pa_State       *tableS;                    /* table of states */
  int            initial;                    /* initial state */
  Bdd_Edge       stateBDD;                   /* BDD of all states */
  int            numTransitions;             /* number of transitions */
  Pa_Transition  *tableT;                    /* table of transitions */
  int            numVar;                     /* number of variables */
  Bdd_Edge       d;                          /* transition relation */
  Bdd_Edge       stab;                       /* stab BDD for process */
} Pa_Process;

typedef struct {
  Bdd_String  name;                       /* composition name */
  int         type;                       /* type of composition */
  int         numProcesses;               /* number of processes */
  int         *tableP;                    /* table of processes */
  int         numActions;                 /* number of given actions */
  int         *tableA;                    /* table of actions */
  int         sort;                       /* sort */
  Bdd_Edge    initialBDD;                 /* BDD of initial state */
  Bdd_Edge    stateBDD;                   /* BDD of all states */
  Bdd_Edge    transitionBDD;              /* BDD of transition relation */
  Bdd_Edge    stab;                       /* stab BDD for composition */
  Bdd_Edge    *di;                        /* di BDDs for processes */
  Bdd_Edge    *dij;                       /* dij BDDs for processes */
  Bdd_Edge    *alphabet;                  /* alphabets of processes */
} Pa_Composition;

/*-----------------------------------------------------------------------*/
/* Variable declarations                                                 */
/*-----------------------------------------------------------------------*/

EXTERN int pa_status;

EXTERN int pa_sorts;                         /* number of sorts */

EXTERN int pa_processes;                     /* number of processes */

EXTERN int pa_compositions;                  /* number of compositions */

EXTERN Pa_Sort *pa_sortTable;                /* table of sorts */

EXTERN Pa_Process *pa_processTable;          /* table of processes */

EXTERN Pa_Composition *pa_compositionTable;  /* table of compositions */

/*-----------------------------------------------------------------------*/
/* Macro declarations                                                    */
/*-----------------------------------------------------------------------*/

/**AutomaticStart*********************************************************/

/*-----------------------------------------------------------------------*/
/* Function prototypes                                                   */
/*-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*/
/* paMain.c                                                              */
/*-----------------------------------------------------------------------*/

EXTERN void Pa_InitPkg();

EXTERN void Pa_ExitPkg();

EXTERN void Pa_AboutPkg();

EXTERN int Pa_AddNewSort(Bdd_String name);

EXTERN int Pa_FindSort(Bdd_String name);

EXTERN int Pa_FOASortAction(Pa_Sort *sort, Bdd_String name);

EXTERN int Pa_FindSortAction(Pa_Sort *sort, Bdd_String name);

EXTERN int Pa_AddNewProcess(Bdd_String name);

EXTERN int Pa_FindProcess(Bdd_String name);

EXTERN int Pa_CopyProcess(Bdd_String name, Bdd_String copy,
                Bdd_String copysort, Bdd_String old, Bdd_String new);

EXTERN Bdd_Boolean Pa_IsGate(Bdd_String name);

EXTERN int Pa_FOAStateProcess(Pa_Process *proc, Bdd_String name);

EXTERN int Pa_FOAGateProcess(Pa_Process *proc, Bdd_String name);

EXTERN int Pa_FOAComposedStateProcess(Pa_Process *proc, Bdd_Edge F,
                                      Bdd_String prefix);

EXTERN int Pa_FindStateProcess(Pa_Process *proc, Bdd_String name);

EXTERN int Pa_AddStateProcess(Pa_Process *proc, Bdd_String name);

EXTERN int Pa_FOANextStateProcess(Pa_Process *proc, Bdd_String prefix);

EXTERN int Pa_NumberStateProcess(Pa_Process *proc);

EXTERN int Pa_FOATransition(Pa_Process *proc, int state1,
                            int action, int type, int state2);

EXTERN int Pa_AddNewComposition(Bdd_String name, int type,
                                int numP, int numA);

EXTERN void Pa_DeleteStateProcess(Pa_Process *proc, int num);

EXTERN void Pa_DeleteSort(int num);

EXTERN void Pa_RemoveDeleteSorts();

EXTERN void Pa_DeleteProcess(int num);

EXTERN void Pa_RemoveDeletedProcesses();

EXTERN void Pa_RemoveComposition(int num);

EXTERN int Pa_FindComposition(Bdd_String name);

EXTERN int Pa_FindProcessComp(Pa_Composition *c, Bdd_String name);

EXTERN int Pa_Process2Composition(Bdd_String name);

EXTERN int Pa_Composition2Process(Bdd_String name, Bdd_Boolean full);

EXTERN void PaSaveSortActions(Pa_Sort *s);

EXTERN void PaSaveProcStates(Pa_Process *p);

/*-----------------------------------------------------------------------*/
/* paKod.c                                                               */
/*-----------------------------------------------------------------------*/

EXTERN void Pa_EncodeSort(Bdd_String name);

EXTERN void Pa_EncodeProcess(Bdd_String name);

EXTERN void Pa_EncodeMultiProcesses(Bdd_String plist);

EXTERN void Pa_ExpandGates(Pa_Process *p);

EXTERN Bdd_String Pa_DecodeSortActions(Pa_Sort *s, Bdd_Edge F, Bdd_Boolean write);

EXTERN void Pa_DecodeProcess(Bdd_String name, Bdd_Boolean write);

EXTERN int Pa_DecodeProcessStates(Pa_Process *p, Bdd_Edge F,
                                   Bdd_Boolean write);

EXTERN int Pa_DecodeProcessTR(Pa_Process *p, Bdd_Edge F,
                               Bdd_Boolean write);

EXTERN void Pa_DecodeComposition(Bdd_String name, Bdd_Boolean write);

EXTERN int Pa_DecodeCompStates(Pa_Composition *c, Bdd_Edge F,
                                Bdd_Boolean write);

EXTERN int Pa_DecodeCompTR(Pa_Composition *c, Bdd_Edge F, Bdd_Boolean write);

EXTERN int Pa_DecodeProcProcPair(Pa_Process *p1, Pa_Process *p2, Bdd_Edge F,
                                  Bdd_Boolean write);

EXTERN int Pa_DecodeCompProcPair(Pa_Composition *c, Pa_Process *p,
                                  Bdd_Edge F, Bdd_Boolean write);

EXTERN int Pa_DecodeCompCompPair(Pa_Composition *c1, Pa_Composition *c2,
                                  Bdd_Edge F, Bdd_Boolean write);

EXTERN Bdd_Edge Pa_ExtractProcTransition(Pa_Process *p, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractCompTransition(Pa_Composition *c, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractProcState(Pa_Process *p, Bdd_Edge f);

EXTERN int Pa_EnumerateProcState(Pa_Process *p, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractCompState(Pa_Composition *c, Bdd_Edge f);

EXTERN Bdd_Edge Pa_ExtractSortAction(Pa_Sort *s, Bdd_Edge f);

EXTERN int Pa_EnumerateSortAction(Pa_Sort *s, Bdd_Edge f);

/*-----------------------------------------------------------------------*/
/* paOutput.c                                                            */
/*-----------------------------------------------------------------------*/

EXTERN void Pa_WriteKnownSorts();

EXTERN void Pa_WriteSort(Bdd_String name);

EXTERN void Pa_WriteKnownProcesses();

EXTERN void Pa_WriteProcess(Bdd_String name, char type);

EXTERN void Pa_WriteKnownCompositions();

EXTERN void Pa_WriteProcessCCS(Bdd_String name, char type);

/*-----------------------------------------------------------------------*/
/* paParser.y                                                            */
/*-----------------------------------------------------------------------*/

EXTERN int Pa_ParseSortProcess(Bdd_String ime);

/**AutomaticEnd***********************************************************/

#endif  /* _PA */

